# coding: utf-8

"""
    OOXML Automation

    This API helps users convert Excel and Powerpoint documents into rich, live dashboards and stories.  # noqa: E501

    The version of the OpenAPI document: 0.1.0-no-tags
    Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

import copy
import logging
import multiprocessing
import sys
import urllib3

import six
from six.moves import http_client as httplib


class TypeWithDefault(type):
    def __init__(cls, name, bases, dct):
        super(TypeWithDefault, cls).__init__(name, bases, dct)
        cls._default = None

    def __call__(cls, **kwargs):
        if cls._default is None:
            cls._default = type.__call__(cls, **kwargs)
        return copy.copy(cls._default)

    def set_default(cls, default):
        cls._default = copy.copy(default)


class Configuration(six.with_metaclass(TypeWithDefault, object)):
    """NOTE: This class is auto generated by OpenAPI Generator

    Ref: https://openapi-generator.tech
    Do not edit the class manually.

    :param host: Base url
    :param api_key: Dict to store API key(s)
    :param api_key_prefix: Dict to store API prefix (e.g. Bearer)
    :param username: Username for HTTP basic authentication
    :param password: Password for HTTP basic authentication
    """

    def __init__(self, host="https://api.presalytics.io/ooxml-automation",
                 api_key={}, api_key_prefix={},
                 username="", password=""):
        """Constructor
        """
        self.host = host
        """Default Base url
        """
        self.temp_folder_path = None
        """Temp file folder for downloading files
        """
        # Authentication Settings
        self.api_key = api_key
        """dict to store API key(s)
        """
        self.api_key_prefix = api_key_prefix
        """dict to store API prefix (e.g. Bearer)
        """
        self.refresh_api_key_hook = None
        """function hook to refresh API key if expired
        """
        self.username = username
        """Username for HTTP basic authentication
        """
        self.password = password
        """Password for HTTP basic authentication
        """
        self.logger = {}
        """Logging Settings
        """
        self.logger["package_logger"] = logging.getLogger("presalytics.client.presalytics_ooxml_automation")
        self.logger["urllib3_logger"] = logging.getLogger("urllib3")
        self.logger_format = '%(asctime)s %(levelname)s %(message)s'
        """Log format
        """
        self.logger_stream_handler = None
        """Log stream handler
        """
        self.logger_file_handler = None
        """Log file handler
        """
        self.logger_file = None
        """Debug file location
        """
        self.debug = False
        """Debug switch
        """

        self.verify_ssl = True
        """SSL/TLS verification
           Set this to false to skip verifying SSL certificate when calling API
           from https server.
        """
        self.ssl_ca_cert = None
        """Set this to customize the certificate file to verify the peer.
        """
        self.cert_file = None
        """client certificate file
        """
        self.key_file = None
        """client key file
        """
        self.assert_hostname = None
        """Set this to True/False to enable/disable SSL hostname verification.
        """

        self.connection_pool_maxsize = multiprocessing.cpu_count() * 5
        """urllib3 connection pool's maximum number of connections saved
           per pool. urllib3 uses 1 connection as default value, but this is
           not the best value when you are making a lot of possibly parallel
           requests to the same host, which is often the case here.
           cpu_count * 5 is used as default value to increase performance.
        """

        self.proxy = None
        """Proxy URL
        """
        self.proxy_headers = None
        """Proxy headers
        """
        self.safe_chars_for_path_param = ''
        """Safe chars for path_param
        """
        self.retries = None
        """Adding retries to override urllib3 default value 3
        """

    @property
    def logger_file(self):
        """The logger file.

        If the logger_file is None, then add stream handler and remove file
        handler. Otherwise, add file handler and remove stream handler.

        :param value: The logger_file path.
        :type: str
        """
        return self.__logger_file

    @logger_file.setter
    def logger_file(self, value):
        """The logger file.

        If the logger_file is None, then add stream handler and remove file
        handler. Otherwise, add file handler and remove stream handler.

        :param value: The logger_file path.
        :type: str
        """
        self.__logger_file = value
        if self.__logger_file:
            # If set logging file,
            # then add file handler and remove stream handler.
            self.logger_file_handler = logging.FileHandler(self.__logger_file)
            self.logger_file_handler.setFormatter(self.logger_formatter)
            for _, logger in six.iteritems(self.logger):
                logger.addHandler(self.logger_file_handler)

    @property
    def debug(self):
        """Debug status

        :param value: The debug status, True or False.
        :type: bool
        """
        return self.__debug

    @debug.setter
    def debug(self, value):
        """Debug status

        :param value: The debug status, True or False.
        :type: bool
        """
        self.__debug = value
        if self.__debug:
            # if debug status is True, turn on debug logging
            for _, logger in six.iteritems(self.logger):
                logger.setLevel(logging.DEBUG)
            # turn on httplib debug
            httplib.HTTPConnection.debuglevel = 1
        else:
            # if debug status is False, turn off debug logging,
            # setting log level to default `logging.WARNING`
            for _, logger in six.iteritems(self.logger):
                logger.setLevel(logging.WARNING)
            # turn off httplib debug
            httplib.HTTPConnection.debuglevel = 0

    @property
    def logger_format(self):
        """The logger format.

        The logger_formatter will be updated when sets logger_format.

        :param value: The format string.
        :type: str
        """
        return self.__logger_format

    @logger_format.setter
    def logger_format(self, value):
        """The logger format.

        The logger_formatter will be updated when sets logger_format.

        :param value: The format string.
        :type: str
        """
        self.__logger_format = value
        self.logger_formatter = logging.Formatter(self.__logger_format)

    def get_api_key_with_prefix(self, identifier):
        """Gets API key (with prefix if set).

        :param identifier: The identifier of apiKey.
        :return: The token for api key authentication.
        """
        if self.refresh_api_key_hook is not None:
            self.refresh_api_key_hook(self)
        key = self.api_key.get(identifier)
        if key:
            prefix = self.api_key_prefix.get(identifier)
            if prefix:
                return "%s %s" % (prefix, key)
            else:
                return key

    def get_basic_auth_token(self):
        """Gets HTTP basic authentication header (string).

        :return: The token for basic HTTP authentication.
        """
        return urllib3.util.make_headers(
            basic_auth=self.username + ':' + self.password
        ).get('authorization')

    def auth_settings(self):
        """Gets Auth Settings dict for api client.

        :return: The Auth Settings information dict.
        """
        return {
        }

    def to_debug_report(self):
        """Gets the essential information for debugging.

        :return: The report for debugging.
        """
        return "Python SDK Debug Report:\n"\
               "OS: {env}\n"\
               "Python Version: {pyversion}\n"\
               "Version of the API: 0.1.0-no-tags\n"\
               "SDK Package Version: 1.0.0".\
               format(env=sys.platform, pyversion=sys.version)

    def get_host_settings(self):
        """Gets an array of host settings

        :return: An array of host settings
        """
        return [
            {
                'url': "https://api.presalytics.io/ooxml-automation",
                'description': "No description provided",
            }
        ]

    def get_host_from_settings(self, index, variables={}):
        """Gets host URL based on the index and variables
        :param index: array index of the host settings
        :param variables: hash of variable and the corresponding value
        :return: URL based on host settings
        """

        servers = self.get_host_settings()

        # check array index out of bound
        if index < 0 or index >= len(servers):
            raise ValueError(
                "Invalid index {} when selecting the host settings. Must be less than {}"  # noqa: E501
                .format(index, len(servers)))

        server = servers[index]
        url = server['url']

        # go through variable and assign a value
        for variable_name in server['variables']:
            if variable_name in variables:
                if variables[variable_name] in server['variables'][
                        variable_name]['enum_values']:
                    url = url.replace("{" + variable_name + "}",
                                      variables[variable_name])
                else:
                    raise ValueError(
                        "The variable `{}` in the host URL has invalid value {}. Must be {}."  # noqa: E501
                        .format(
                            variable_name, variables[variable_name],
                            server['variables'][variable_name]['enum_values']))
            else:
                # use default value
                url = url.replace(
                    "{" + variable_name + "}",
                    server['variables'][variable_name]['default_value'])

        return url



import os
from environs import Env

env = Env()
env.read_env()

host = env("STORY_HOST", "https://api.presalytics.io")

SPEC = {
        "update": True,
        "update_type": "patch",
        "endpoint": "{0}/story/no_tags_spec".format(host),
        "package_name": "presalytics_story",
        "package_url": "https://github.com/presalytics/story-python-client.git",
        "readme_path": os.path.join(os.path.dirname(os.path.realpath(__file__)), "README.md"),
        "setuppy_path": os.path.join(os.path.dirname(os.path.realpath(__file__)), "setup.py"),
        "pypi_url": "https://pypi.org/project/presalytics-story/"
}


import requests, os, shutil, importlib.util, fileinput
from io import BytesIO
from zipfile import ZipFile
from pprint import pprint
from setuptools import sandbox
from environs import Env
from tempfile import mkstemp
from shutil import move
from os import fdopen, remove
from doc_converter.spec import SPEC as doc_converter_spec
from ooxml_automation.spec import SPEC as ooxml_automation_spec
from story.spec import SPEC as story_spec

DELETE_TMP_FILES = True

env = Env()
env.read_env()

CODEGEN_ENDPOINT="https://openapi-generator.presalytics.io/api/gen/clients/python"
CODEGEN_DL_STUB="https://openapi-generator.presalytics.io/api/gen/download/"
CODEGEN_DIR=os.path.dirname(os.path.realpath(__file__))
TMP_PATH = os.path.join(CODEGEN_DIR, "tmp")
LIC_PATH = os.path.join(CODEGEN_DIR, "LICENSE")


CLIENT_SPECS = []
CLIENT_SPECS.append(doc_converter_spec)
CLIENT_SPECS.append(ooxml_automation_spec)
CLIENT_SPECS.append(story_spec)
GIT_SSH_COMMAND = 'ssh -i ~/.ssh/id_rsa'
os.environ['GIT_SSH_COMMAND'] = GIT_SSH_COMMAND

os.system("git config credential.helper presalytics-bot")

HEADER = {
    'Content-type': 'application/json',
    'Accept': 'application/json'
}


# from git import Repo

# PATH_OF_GIT_REPO = r'path\to\your\project\folder\.git'  # make sure .git folder is properly configured
# COMMIT_MESSAGE = 'comment from python script'



def increment_version(old_version, update_type):

    _major, _minor, _patch = old_version.split(".")
    if update_type == "major":
        _major = int(_major) + 1
        _minor = 0
        _patch = 0
    elif update_type == "minor":
        _minor = int(_minor) + 1
        _patch = 0
    else:
        _patch = int(_patch) + 1
    new_version = "{0}.{1}.{2}".format(_major, _minor, _patch)
    return new_version

def replace_ver(file_path, update_type):
    #Create temp file
    fh, abs_path = mkstemp()
    with fdopen(fh,'w') as new_file:
        with open(file_path) as old_file:
            for line in old_file:
                if line[:10] == "VERSION = ":
                    ver = line.replace("VERSION = ", '').replace('"','').replace('\n','')
                    new_ver = increment_version(ver, update_type)
                    new_file.write('VERSION = "{0}"\n'.format(new_ver))
                else:
                    new_file.write(line)

    #Remove original file
    remove(file_path)
    #Move new file
    move(abs_path, file_path)
    return new_ver


def update_imports(tmp_dir_path, package_name):
    new_import_path = "presalytics.client.{}".format(package_name)
    find1 = "from {}".format(package_name)
    replace1 = "from {}".format(new_import_path)
    find2 = "import {}".format(package_name)
    replace2 = "import {}".format(new_import_path)
    find3 = "klass = getattr({}.models, klass)".format(package_name)
    replace3 = "klass = getattr({}.models, klass)".format(new_import_path)
    find4 = 'logging.getLogger(\"{}'.format(package_name)
    replace4 = 'logging.getLogger(\"{}'.format(new_import_path)

    for dname, dirs, files in os.walk(tmp_dir_path):
        for fname in files:
            fpath = os.path.join(dname, fname)
            with open(fpath) as f:
                s = f.read()
            s = s.replace(find1, replace1)
            s = s.replace(find2, replace2)
            s = s.replace(find3, replace3)
            s = s.replace(find4, replace4)
            with open(fpath, "w") as f:
                f.write(s)

for spec in CLIENT_SPECS:
    try:
        if spec["update"] is True:
            #new_ver = replace_ver(spec["setuppy_path"], spec["update_type"])
            payload = {
                'openAPIUrl': spec['endpoint'],
                'options' : {
                    'packageName': spec['package_name'],
                    'projectName': 'Presalytics API'
                    # 'packageVersion': new_ver
                }
            }

            print("Sending {} api spec to code generator".format(spec['package_name']))
            response = requests.post(CODEGEN_ENDPOINT, json=payload, headers=HEADER)
            print("Response from openapi generator:")
            pprint(response)
            uid = response.json()['code']
            link = CODEGEN_DL_STUB + uid

            print("Downloading generated code for {} ".format(spec['package_name']))
            file_response = requests.get(link)

            zipfile = ZipFile(BytesIO(file_response.content))
            zipfile.extractall(TMP_PATH)
            update_imports(TMP_PATH, spec["package_name"])
            source_dir = os.path.join(os.path.dirname(__file__), "tmp", "python-client", spec["package_name"])
            root = os.path.dirname(os.path.dirname(__file__))
            target_dir = os.path.join(root, "presalytics", "client", spec["package_name"])
            try:
                shutil.rmtree(target_dir)
            except Exception:
                pass
            shutil.copytree(source_dir, target_dir)
            print("Subpackage {0} successfull updated in {1}".format(spec["package_name"], target_dir))
    except Exception:
        print("Sub package generation unsuccessful.  Please debug and retry")
    finally:
        if DELETE_TMP_FILES:
            try:
                shutil.rmtree(TMP_PATH)
            except Exception:
                pass

    # repo = git.get_repo("presalytics/" + spec['package_name'])

#Clean temp files



# coding: utf-8

"""
    Story

    This API is the main entry point for creating, editing and publishing analytics throught the Presalytics API  # noqa: E501

    The version of the OpenAPI document: 0.3.1
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six


class StoryOutlineHistory(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'created_at': 'datetime',
        'created_by': 'str',
        'id': 'str',
        'updated_at': 'datetime',
        'updated_by': 'str',
        'collaborator_user_id': 'str',
        'outline': 'str',
        'revision_number': 'int',
        'story_id': 'str'
    }

    attribute_map = {
        'created_at': 'created_at',
        'created_by': 'created_by',
        'id': 'id',
        'updated_at': 'updated_at',
        'updated_by': 'updated_by',
        'collaborator_user_id': 'collaborator_user_id',
        'outline': 'outline',
        'revision_number': 'revision_number',
        'story_id': 'story_id'
    }

    def __init__(self, created_at=None, created_by=None, id=None, updated_at=None, updated_by=None, collaborator_user_id=None, outline=None, revision_number=None, story_id=None):  # noqa: E501
        """StoryOutlineHistory - a model defined in OpenAPI"""  # noqa: E501

        self._created_at = None
        self._created_by = None
        self._id = None
        self._updated_at = None
        self._updated_by = None
        self._collaborator_user_id = None
        self._outline = None
        self._revision_number = None
        self._story_id = None
        self.discriminator = None

        if created_at is not None:
            self.created_at = created_at
        if created_by is not None:
            self.created_by = created_by
        if id is not None:
            self.id = id
        if updated_at is not None:
            self.updated_at = updated_at
        if updated_by is not None:
            self.updated_by = updated_by
        if collaborator_user_id is not None:
            self.collaborator_user_id = collaborator_user_id
        if outline is not None:
            self.outline = outline
        if revision_number is not None:
            self.revision_number = revision_number
        if story_id is not None:
            self.story_id = story_id

    @property
    def created_at(self):
        """Gets the created_at of this StoryOutlineHistory.  # noqa: E501


        :return: The created_at of this StoryOutlineHistory.  # noqa: E501
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this StoryOutlineHistory.


        :param created_at: The created_at of this StoryOutlineHistory.  # noqa: E501
        :type: datetime
        """

        self._created_at = created_at

    @property
    def created_by(self):
        """Gets the created_by of this StoryOutlineHistory.  # noqa: E501


        :return: The created_by of this StoryOutlineHistory.  # noqa: E501
        :rtype: str
        """
        return self._created_by

    @created_by.setter
    def created_by(self, created_by):
        """Sets the created_by of this StoryOutlineHistory.


        :param created_by: The created_by of this StoryOutlineHistory.  # noqa: E501
        :type: str
        """

        self._created_by = created_by

    @property
    def id(self):
        """Gets the id of this StoryOutlineHistory.  # noqa: E501


        :return: The id of this StoryOutlineHistory.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this StoryOutlineHistory.


        :param id: The id of this StoryOutlineHistory.  # noqa: E501
        :type: str
        """

        self._id = id

    @property
    def updated_at(self):
        """Gets the updated_at of this StoryOutlineHistory.  # noqa: E501


        :return: The updated_at of this StoryOutlineHistory.  # noqa: E501
        :rtype: datetime
        """
        return self._updated_at

    @updated_at.setter
    def updated_at(self, updated_at):
        """Sets the updated_at of this StoryOutlineHistory.


        :param updated_at: The updated_at of this StoryOutlineHistory.  # noqa: E501
        :type: datetime
        """

        self._updated_at = updated_at

    @property
    def updated_by(self):
        """Gets the updated_by of this StoryOutlineHistory.  # noqa: E501


        :return: The updated_by of this StoryOutlineHistory.  # noqa: E501
        :rtype: str
        """
        return self._updated_by

    @updated_by.setter
    def updated_by(self, updated_by):
        """Sets the updated_by of this StoryOutlineHistory.


        :param updated_by: The updated_by of this StoryOutlineHistory.  # noqa: E501
        :type: str
        """

        self._updated_by = updated_by

    @property
    def collaborator_user_id(self):
        """Gets the collaborator_user_id of this StoryOutlineHistory.  # noqa: E501


        :return: The collaborator_user_id of this StoryOutlineHistory.  # noqa: E501
        :rtype: str
        """
        return self._collaborator_user_id

    @collaborator_user_id.setter
    def collaborator_user_id(self, collaborator_user_id):
        """Sets the collaborator_user_id of this StoryOutlineHistory.


        :param collaborator_user_id: The collaborator_user_id of this StoryOutlineHistory.  # noqa: E501
        :type: str
        """

        self._collaborator_user_id = collaborator_user_id

    @property
    def outline(self):
        """Gets the outline of this StoryOutlineHistory.  # noqa: E501


        :return: The outline of this StoryOutlineHistory.  # noqa: E501
        :rtype: str
        """
        return self._outline

    @outline.setter
    def outline(self, outline):
        """Sets the outline of this StoryOutlineHistory.


        :param outline: The outline of this StoryOutlineHistory.  # noqa: E501
        :type: str
        """

        self._outline = outline

    @property
    def revision_number(self):
        """Gets the revision_number of this StoryOutlineHistory.  # noqa: E501


        :return: The revision_number of this StoryOutlineHistory.  # noqa: E501
        :rtype: int
        """
        return self._revision_number

    @revision_number.setter
    def revision_number(self, revision_number):
        """Sets the revision_number of this StoryOutlineHistory.


        :param revision_number: The revision_number of this StoryOutlineHistory.  # noqa: E501
        :type: int
        """

        self._revision_number = revision_number

    @property
    def story_id(self):
        """Gets the story_id of this StoryOutlineHistory.  # noqa: E501


        :return: The story_id of this StoryOutlineHistory.  # noqa: E501
        :rtype: str
        """
        return self._story_id

    @story_id.setter
    def story_id(self, story_id):
        """Sets the story_id of this StoryOutlineHistory.


        :param story_id: The story_id of this StoryOutlineHistory.  # noqa: E501
        :type: str
        """

        self._story_id = story_id

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, StoryOutlineHistory):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



# coding: utf-8

"""
    Story

    This API is the main entry point for creating, editing and publishing analytics throught the Presalytics API  # noqa: E501

    The version of the OpenAPI document: 0.3.1
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six


class StoryCollaboratorAllOf(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'active': 'bool',
        'email': 'str',
        'lead_id': 'int',
        'name': 'str',
        'permission_type': 'PermissionType',
        'permission_type_id': 'str',
        'story_id': 'str',
        'user_id': 'str'
    }

    attribute_map = {
        'active': 'active',
        'email': 'email',
        'lead_id': 'lead_id',
        'name': 'name',
        'permission_type': 'permission_type',
        'permission_type_id': 'permission_type_id',
        'story_id': 'story_id',
        'user_id': 'user_id'
    }

    def __init__(self, active=None, email=None, lead_id=None, name=None, permission_type=None, permission_type_id=None, story_id=None, user_id=None):  # noqa: E501
        """StoryCollaboratorAllOf - a model defined in OpenAPI"""  # noqa: E501

        self._active = None
        self._email = None
        self._lead_id = None
        self._name = None
        self._permission_type = None
        self._permission_type_id = None
        self._story_id = None
        self._user_id = None
        self.discriminator = None

        if active is not None:
            self.active = active
        if email is not None:
            self.email = email
        if lead_id is not None:
            self.lead_id = lead_id
        if name is not None:
            self.name = name
        if permission_type is not None:
            self.permission_type = permission_type
        if permission_type_id is not None:
            self.permission_type_id = permission_type_id
        if story_id is not None:
            self.story_id = story_id
        if user_id is not None:
            self.user_id = user_id

    @property
    def active(self):
        """Gets the active of this StoryCollaboratorAllOf.  # noqa: E501


        :return: The active of this StoryCollaboratorAllOf.  # noqa: E501
        :rtype: bool
        """
        return self._active

    @active.setter
    def active(self, active):
        """Sets the active of this StoryCollaboratorAllOf.


        :param active: The active of this StoryCollaboratorAllOf.  # noqa: E501
        :type: bool
        """

        self._active = active

    @property
    def email(self):
        """Gets the email of this StoryCollaboratorAllOf.  # noqa: E501


        :return: The email of this StoryCollaboratorAllOf.  # noqa: E501
        :rtype: str
        """
        return self._email

    @email.setter
    def email(self, email):
        """Sets the email of this StoryCollaboratorAllOf.


        :param email: The email of this StoryCollaboratorAllOf.  # noqa: E501
        :type: str
        """

        self._email = email

    @property
    def lead_id(self):
        """Gets the lead_id of this StoryCollaboratorAllOf.  # noqa: E501


        :return: The lead_id of this StoryCollaboratorAllOf.  # noqa: E501
        :rtype: int
        """
        return self._lead_id

    @lead_id.setter
    def lead_id(self, lead_id):
        """Sets the lead_id of this StoryCollaboratorAllOf.


        :param lead_id: The lead_id of this StoryCollaboratorAllOf.  # noqa: E501
        :type: int
        """

        self._lead_id = lead_id

    @property
    def name(self):
        """Gets the name of this StoryCollaboratorAllOf.  # noqa: E501


        :return: The name of this StoryCollaboratorAllOf.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this StoryCollaboratorAllOf.


        :param name: The name of this StoryCollaboratorAllOf.  # noqa: E501
        :type: str
        """

        self._name = name

    @property
    def permission_type(self):
        """Gets the permission_type of this StoryCollaboratorAllOf.  # noqa: E501


        :return: The permission_type of this StoryCollaboratorAllOf.  # noqa: E501
        :rtype: PermissionType
        """
        return self._permission_type

    @permission_type.setter
    def permission_type(self, permission_type):
        """Sets the permission_type of this StoryCollaboratorAllOf.


        :param permission_type: The permission_type of this StoryCollaboratorAllOf.  # noqa: E501
        :type: PermissionType
        """

        self._permission_type = permission_type

    @property
    def permission_type_id(self):
        """Gets the permission_type_id of this StoryCollaboratorAllOf.  # noqa: E501


        :return: The permission_type_id of this StoryCollaboratorAllOf.  # noqa: E501
        :rtype: str
        """
        return self._permission_type_id

    @permission_type_id.setter
    def permission_type_id(self, permission_type_id):
        """Sets the permission_type_id of this StoryCollaboratorAllOf.


        :param permission_type_id: The permission_type_id of this StoryCollaboratorAllOf.  # noqa: E501
        :type: str
        """

        self._permission_type_id = permission_type_id

    @property
    def story_id(self):
        """Gets the story_id of this StoryCollaboratorAllOf.  # noqa: E501


        :return: The story_id of this StoryCollaboratorAllOf.  # noqa: E501
        :rtype: str
        """
        return self._story_id

    @story_id.setter
    def story_id(self, story_id):
        """Sets the story_id of this StoryCollaboratorAllOf.


        :param story_id: The story_id of this StoryCollaboratorAllOf.  # noqa: E501
        :type: str
        """

        self._story_id = story_id

    @property
    def user_id(self):
        """Gets the user_id of this StoryCollaboratorAllOf.  # noqa: E501


        :return: The user_id of this StoryCollaboratorAllOf.  # noqa: E501
        :rtype: str
        """
        return self._user_id

    @user_id.setter
    def user_id(self, user_id):
        """Sets the user_id of this StoryCollaboratorAllOf.


        :param user_id: The user_id of this StoryCollaboratorAllOf.  # noqa: E501
        :type: str
        """

        self._user_id = user_id

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, StoryCollaboratorAllOf):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



""" Constants for api.presalytics.io client """
HOST = "https://api.presalytics.io"
SITE_HOST = "https://presalytics.io"
OIDC_WELL_KNOWN = "https://auth.presalytics.io/auth/realms/presalytics/.well-known/openid-configuration"
OIDC_TOKEN = "https://auth.presalytics.io/auth/realms/presalytics/protocol/openid-connect/token"
PORT = 443
OIDC_AUTH_HOST = "https://login.presalytics.io/"
OIDC_AUTH_PORT = 443
REDIRECT_URI = "https://presalytics.io/user/login-success"
OIDC_REALM = "presalytics"
DEFAULT_CLIENT_ID = "JdypLkpW4JF7ORdNnfcul5N4mifzS5MC"
TOKEN_FILE = "token.json"
LOCALHOST_SERVER = "127.0.0.1"
LOCALHOST_PORT = 8052
LOGIN_PATH = "/accounts/login/"
API_LOGGEDIN_NEXT_URL = "/user/api-logged-in/"
STORY_VIEW_URL = "/story/view/{0}/"
STORY_MANAGE_URL = "/story/manage/{0}/"
DEFAULT_AUDIENCE = "https://api.presalytics.io/"


import os
from environs import Env

env = Env()
env.read_env()

doc_converter_host = env("DOC_CONVERTER_HOST", "https://api.presalytics.io")

SPEC = {
        "update": False,
        "update_type": "patch",
        "endpoint": "{}/doc-converter/openapi.json".format(doc_converter_host),
        "package_name": "presalytics_doc_converter",
        "package_url": "https://github.com/presalytics/doc-converter-python-client.git",
        "readme_path": os.path.join(os.path.dirname(os.path.realpath(__file__)), "README.md"),
        "setuppy_path": os.path.join(os.path.dirname(os.path.realpath(__file__)), "setup.py"),
        "pypi_url": "https://pypi.org/project/presalytics-doc-converter/"
}


import mpld3
import jinja2
import typing
import json
import lxml
import os
import presalytics.story.components
import presalytics.story.outline
import presalytics.lib.exceptions
import presalytics.lib.constants
import presalytics.lib.plugins.matplotlib
import presalytics.lib.plugins.external
if typing.TYPE_CHECKING:
    from matplotlib.pyplot import Figure


class MatplotlibFigure(presalytics.story.components.WidgetBase):
    """
    A `Widget` for rendering `matplotlib.pyplot.Figure` instances in stories

    This class acts as wrapper class for matplotlib figures, allowing their packaging into
    `presalytics.story.outline.Widget` objects and serialization to json.  At render-time, 
    the figure is converted to a d3.js object via the [mpld3](https://mpld3.github.io/)
    package.

    Parameters
    ----------

    figure : `matplotlib.pyplot.Figure`
        the figure to create the widget from

    name : str
        the name of widget.  Must be unique within `presalytics.COMPONENTS`

    Attributes
    ----------
    figure_dict : dict
        A `dict` containing json-serializable data for reconstituting the
        `matplotlib.pyplot.Figure` object.

    figure_id : str
        A unique identifier used to render the that figure into a d3.js object
    """
    __component_kind__ = 'matplotlib-figure'
    additional_properties: typing.Dict

    def __init__(self, figure: 'Figure', name: str, *args, **kwargs):
        super().__init__(self, *args, **kwargs)
        self.fig = figure
        self.name = name
        self.figure_dict = None
        if self.fig:
            self.figure_dict = mpld3.fig_to_dict(figure)
        else:
            figure_dict = kwargs.pop("figure_dict", None)
            if figure_dict:
                self.figure_dict = figure_dict
        if self.figure_dict is None:
            message = "MatplotlibFigure requires a figure_dict attribute.  Please supply either a valid figure_dict or matplotlib.pyplot.Figure object to __init__"
            raise presalytics.lib.exceptions.ValidationError(message)
        self.figure_id = self.figure_dict["id"]
        self.additional_properties = {}
        for key, val in kwargs.items():
            self.additional_properties[key] = val
        self.outline_widget = self.serialize()

    def to_html(self):
        return '<div id="{0}" class="mpld3"></div>'.format(self.figure_id)

    @classmethod
    def deserialize(cls, outline, **kwargs):
        figure_dict = outline.data.get("figure_dict")
        kwargs.update(**outline.additional_properties)
        return cls(None, outline.name, figure_dict=figure_dict, **outline.additional_properties)

    def serialize(self, **kwargs):
        data = {
            'figure_dict': self.figure_dict,
            'id': self.figure_id
        }
        plugin_obj = presalytics.story.outline.Plugin(
            name='mpld3',
            kind='script',
            config=data
        )
        return presalytics.story.outline.Widget(
            name=self.name,
            kind=self.__component_kind__,
            plugins=[plugin_obj.to_dict()],
            data=data,
            additional_properties=self.additional_properties
        )


class MatplotlibResponsiveFigure(MatplotlibFigure):
    """
    A `Widget` for rendering `matplotlib.pyplot.Figure` instances in stories

    This class acts as wrapper class for matplotlib figures, allowing their packaging into
    `presalytics.story.outline.Widget` objects and serialization to json.  At render-time, 
    the figure is converted to a d3.js object via the [mpld3](https://mpld3.github.io/)
    package.

    This class also depends on the static plugin

    Parameters
    ----------

    figure : `matplotlib.pyplot.Figure`
        the figure to create the widget from

    name : str
        the name of widget.  Must be unique within `presalytics.COMPONENTS`

    story_id : str, optional
        The story of thet

    Attributes
    ----------
    figure_dict : dict
        A `dict` containing json-serializable data for reconstituting the
        `matplotlib.pyplot.Figure` object.

    figure_id : str
        A unique identifier used to render the that figure into a d3.js object
    """

    additional_properties: typing.Dict
    __component_kind__ = 'matplotlib-responsive'

    __plugins__ = [
        {
            'name': 'external_scripts',
            'kind': 'script',
            'config': {
                'approved_scripts_key': 'jquery'
            }
        },
        {
            'name': 'external_scripts',
            'kind': 'script',
            'config': {
                'approved_scripts_key': 'mpl-responsive'
            }
        },
        {
            'name': 'external_links',
            'kind': 'style',
            'config': {
                'approved_styles_key': 'preloaders'
            }
        }
    ] #type: ignore

    def __init__(self, figure: 'Figure', name: str, story_id: str = "empty", *args, **kwargs):
        self.story_id = story_id
        super(MatplotlibResponsiveFigure, self).__init__(figure, name, *args, **kwargs)
        self.story_host = self.get_client(delegate_login=True).story.api_client.configuration.host

        
    def to_html(self):
        if not self.story_id:
            message = "This object requires a valid story_id to render."
            raise presalytics.lib.exceptions.MissingConfigException(message=message)
        html = self.create_container()
        return html


    def create_container(self, **kwargs):
        """
        Wraps the Matplotlib Figure in a SVG endpoint load via `<iframe>` that
        will be rendered inside of a story and rescaled to give repsonsive effect
        """
        params = {
            "site_host": self.get_client(delegate_login=True).site_host,
            "figure_id": self.figure_id,
            "story_id": self.story_id
        }
        source_url = "{site_host}/story/matplotlib-figure/{story_id}/{figure_id}".format(**params)
        empty_parent_div = lxml.html.Element("div", {
            'class': 'empty-parent bg-light matplotlib-responsive-container',
            'style': 'height: 100%; width: 100%, display: block; text-align: left;'
        })
        frame = lxml.html.Element("iframe", {
            'src': source_url,
            'frameborder': "0",
            'scrolling': "no",
            'class': 'matplotlib-responsive-frame',
            'style': 'max-height: none; max-width: none;'
        })
        empty_parent_div.append(frame)
        return lxml.html.tostring(empty_parent_div).decode('utf-8')


    @classmethod
    def deserialize(cls, outline, **kwargs):
        figure_dict = outline.data.get("figure_dict")
        story_id = outline.data.get("story_id", None)
        kwargs.update(**outline.additional_properties)
        return cls(None, 
                   outline.name,
                   story_id,
                   figure_dict=figure_dict,
                   **kwargs)

    def serialize(self, **kwargs):
        data = {
            'figure_dict': self.figure_dict,
            'id': self.figure_id,
            'story_id': self.story_id
        }
        return presalytics.story.outline.Widget(
            name=self.name,
            kind=self.__component_kind__,
            data=data,
            additional_properties=self.additional_properties
        )

    def standalone_html(self) -> str:
        SIMPLE_HTML = jinja2.Template("""<!DOCTYPE html>
        <html>
            <head>
                <meta charset="UTF-8">
                <style>
                {{ extra_css }}
                </style>
            </head>
            <body>
                <script type="text/javascript" src="{{ d3_url }}"></script>
                <script type="text/javascript" src="{{ mpld3_url }}"></script>
                <div id="{{ figid }}"></div>
                <script type="text/javascript">

                    mpld3.draw_figure("{{ figid }}", {{ figure_json }});
            
                </script>
            </body>
        </html>""")
        figure_json = json.dumps(self.figure_dict, cls=presalytics.lib.plugins.matplotlib.Mpld3NumpyToJson)
        context = {
            "d3_url": presalytics.lib.plugins.external.ApprovedExternalScripts().attr_dict.flatten().get('d3v3'),
            "mpld3_url": presalytics.lib.plugins.external.ApprovedExternalScripts().attr_dict.flatten().get('mpld3'),
            "figid": self.figure_id,
            "figure_json": figure_json
        }
        return SIMPLE_HTML.render(**context)


import typing
import presalytics
import presalytics.client.auth
import presalytics.story.components
import presalytics.story.outline
import presalytics.story.util
import presalytics.client.presalytics_ooxml_automation.models


class OoxmlTheme(presalytics.story.components.ThemeBase):
    """
    TODO: Review whether this class is obsolete
    """
    name: str
    ooxml_id: str
    plugin_config: typing.Dict
    always_refresh: bool

    __component_kind__ = 'ooxml-theme'

    def __init__(self, 
                 name, 
                 ooxml_theme_id, 
                 plugin_config=None, 
                 always_refresh=False,
                 client_info={},
                 **kwargs):
        super(OoxmlTheme, self).__init__(**kwargs)
        self.name = name
        self.ooxml_id = ooxml_theme_id
        self.always_refresh = always_refresh
        self.client_kwargs = client_info
        if not plugin_config or self.always_refresh:
            if not plugin_config:
                self.plugin_config = {}
            self.get_configuration()

    def get_configuration(self):
        client = presalytics.Client(**self.client_kwargs)
        theme = client.ooxml_automation.theme_themes_details_get_id(self.ooxml_id)
        extra_params = ['dateCreated', 'dateModified', 'userCreated', 'userModified', 'id', 'themeId']
        colors = {k: v for k, v in theme.colors.to_dict().items() if k not in extra_params}
        fonts = {k: v for k, v in theme.fonts.to_dict().items() if k not in extra_params}
        slide_details = client.ooxml_automation.slides_slides_details_get_id(theme.slide_id)
        color_map_dict = slide_details.slide_master.to_dict()["color_map"]
        color_types = client.ooxml_automation.shared_colortypes_get()

        mapped_colors = {
            "background1": self.map_color_type("background1", color_map_dict, colors, color_types),
            "background2": self.map_color_type("background2", color_map_dict, colors, color_types),
            "text1": self.map_color_type("text1", color_map_dict, colors, color_types),
            "text2": self.map_color_type("text2", color_map_dict, colors, color_types)
        }
        color_params = {k: v for k, v in colors.items() if k not in extra_params}
        color_params.update(mapped_colors)
        params = {k: v for k, v in color_params.items() if k not in extra_params}
        params.update(fonts)

        self.plugin_config = params
    

    def map_color_type(
            self,
            color_map_name: str,
            color_map: typing.Dict,
            theme_colors: typing.Dict,
            color_types_list=None) -> str:
        if not color_types_list:
            client = presalytics.Client(**self.client_kwargs)
            color_types_list = client.ooxml_automation.shared_colortypes_get()
        color_id = color_map[color_map_name]
        color_name = next(x.name for x in color_types_list if x.type_id == color_id)
        key = color_name[0].lower() + color_name[1:]
        color = theme_colors.get(key, None)
        return color

    def serialize(self):
        plugin = presalytics.story.outline.Plugin(
            kind='style',
            name='ooxml-theme',
            config=self.plugin_config
        )
        data = {
            "ooxml_theme_id": self.ooxml_id,
            "always_refresh": self.always_refresh,
        }
        theme = presalytics.story.outline.Theme(
            kind=self.__component_kind__,
            name=self.name,
            data=data,
            plugins=[plugin]
        )
        return theme

    @classmethod
    def deseriailize(cls, component, **kwargs):
        plugin_config = component.plugins[0].config
        return cls(
            name=component.name,
            ooxml_theme_id=component.data["ooxml_theme_id"],
            plugin_config=plugin_config,
            always_refresh=component.data["always_refresh"],
            **kwargs
        )



import typing
import presalytics
import presalytics.lib.plugins.base
import presalytics.lib.exceptions

site_host = "https://presalytics.io"
try:
    site_host = presalytics.CONFIG["HOSTS"]["SITE"] #type: ignore
except (KeyError, AttributeError, ImportError, ModuleNotFoundError):
    pass


class AttrDict(dict):
    """ Dictionary subclass whose entries can be accessed by attributes
        (as well as normally).
    """
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self

    @staticmethod
    def from_nested_dict(data):
        """ Construct nested AttrDicts from nested dictionaries. """
        if not isinstance(data, dict):
            return data
        else:
            return AttrDict({key: AttrDict.from_nested_dict(data[key])
                            for key in data})

    def flatten(self) -> typing.Dict:
        return AttrDict._flatten(self, '', {})

    @staticmethod
    def _flatten(current, key, result) -> typing.Dict:
        if isinstance(current, dict):
            for k in current:
                new_key = "{0}.{1}".format(key, k) if len(key) > 0 else k
                AttrDict._flatten(current[k], new_key, result)
        else:
            result[key] = current
        return result


class ApprovedExternalLinks(presalytics.lib.plugins.base.StylePlugin):
    """
    `presalytics.lib.plugins.base.StylePlugin` subclass for converting a `presalytics.story.outline.Plugin` 
    config into an html `<link>` fragment.

    Attributes
    ----------
    attr_dict: presalytics.lib.plugins.external.AttrDict
        Performs nested lookups on the `STYLES_MAP`
    """
    __plugin_name__ = 'external_links'

    STYLES_MAP = {
        'reveal': {
            'base': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/css/reveal.min.css',
            'themes': {
                'beige': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/css/theme/beige.min.css',
                'black': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/css/theme/black.min.css',
                'blood': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/css/theme/blood.min.css',
                'league': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/css/theme/league.min.css',
                'moon': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/css/theme/moon.min.css',
                'night': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/css/theme/night.min.css',
                'serif': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/css/theme/serif.min.css',
                'simple': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/css/theme/simple.min.css',
                'sky': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/css/theme/sky.min.css',
                'solarized': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/css/theme/solarized.min.css',
                'white': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/css/theme/white.min.css'
            },
            'print': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/css/print/pdf.min.css',
            'customizations': '{0}/static/css/reveal-customizations.css'.format(site_host) 
        },
        'preloaders' : '{0}/static/css/preloaders.css'.format(site_host),
        'bootstrap4': "https://stackpath.bootstrapcdn.com/bootstrap/4.4.1/css/bootstrap.min.css"
    }
    
    """
    Static nested dictionary containing links to external stylesheets that will be rendered alongside this plugin
    """

    def __init__(self, **kwargs):
        super(ApprovedExternalLinks, self).__init__(**kwargs)
        self.attr_dict = AttrDict(self.STYLES_MAP)


    def to_style(self, config, **kwargs):
        """
        Converts a dot-notation key for nested dictionaries (e.g., 'reveal.base') into a
        string contain an html fragement with `<link>` tag.  The dot-notation key 
        is pulled from the 'approved_styles_key' of the 'config' attrubte of subclass 
        of a `presalytics.story.outline.Plugin` object.
        """
        key = config['approved_styles_key']
        link = self.attr_dict.flatten()[key]
        if link is None:
            message = "Key {0} does not reference a link in the APPROVED_STYLES dictionary".format(key)
            raise presalytics.lib.exceptions.MissingConfigException(message)
        return '<link rel="stylesheet" href="{0}"/>'.format(link)


class ApprovedExternalScripts(presalytics.lib.plugins.base.ScriptPlugin):
    """
    `presalytics.lib.plugins.base.ScriptPlugin` subclass for converting a `presalytics.story.outline.Plugin` 
    config into an html `<script>` fragment.

    Attributes
    ----------
    attr_dict: presalytics.lib.plugins.external.AttrDict
        Performs nested lookups on the `SCRIPT_MAP`
    """
    __plugin_name__ = 'external_scripts'

    SCRIPT_MAP = {
        'd3': 'https://d3js.org/d3.v5.min.js',
        'd3v3': 'https://cdnjs.cloudflare.com/ajax/libs/d3/3.5.17/d3.min.js',
        'reveal': {
            'base': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/js/reveal.min.js',
            'marked': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/plugin/markdown/marked.js',
            'markdown': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/plugin/markdown/markdown.min.js',
            'highlight': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/plugin/highlight/highlight.min.js',
            'math': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/plugin/math/math.min.js',
            'zoom': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/plugin/zoom-js/zoom.min.js',
            'notes': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/plugin/notes/notes.min.js',
            'print': 'https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.6.0/plugin/print-pdf/print-pdf.min.js',
            'customizations': '{0}/static/js/revealcustomizations.js'.format(site_host),

        },
        'mpld3': '{0}/static/mpld3/mpld3.min.js'.format(site_host),
        'ooxml': '{0}/static/ooxml/ooxml.js'.format(site_host),
        'mpl-responsive': '{0}/static/js/mpl-responsive.js'.format(site_host),
        'jquery': 'https://code.jquery.com/jquery-3.4.1.min.js',
        'popper': 'https://cdnjs.cloudflare.com/ajax/libs/popper.js/1.14.3/umd/popper.min.js',
        'bootstrap4': 'https://stackpath.bootstrapcdn.com/bootstrap/4.4.1/js/bootstrap.min.js'
    }
    """
    Static nested dictionary containing links to external scripts that will be rendered alongside this plugin
    """

    def __init__(self, **kwargs):
        super(ApprovedExternalScripts, self).__init__(**kwargs)
        self.attr_dict = AttrDict(self.SCRIPT_MAP)


    def to_script(self, config, **kwargs):
        """
        Converts a dot-notation key for nested dictionaries (e.g., 'reveal.base') into a
        string containing an html fragement with `<script>` tags.  The dot-notation key 
        is pulled from the 'approved_styles_key' of the 'config' attrubte of subclass 
        of a `presalytics.story.outline.Plugin` object.
        """
        key = config['approved_scripts_key']
        link = self.attr_dict.flatten()[key]
        if link is None:
            message = "Key {0} does not reference a link in the APPROVED_SCRIPTS dictionary".format(key)
            raise presalytics.lib.exceptions.MissingConfigException(message)
        return '<script type="text/javascript" src="{0}"></script>'.format(link)



import os
import shutil
import lxml.html
import webbrowser
import logging
import collections
import typing
import tempfile
import lxml
import lxml.html
import lxml.etree
import lxml.html.builder as E
import datetime
import six
import sys
import urllib.parse
import presalytics
import presalytics.lib
import presalytics.lib.plugins
import presalytics.lib.plugins.base
import presalytics.lib.plugins.reveal
import presalytics.lib.templates
import presalytics.lib.templates.base
import presalytics.story.util
import presalytics.story.server
import presalytics.story.components
import presalytics.lib.exceptions
if typing.TYPE_CHECKING:
    from presalytics.story.outline import StoryOutline, Page, Plugin
    from presalytics.story.components import PageTemplateBase, ComponentBase, WidgetBase, ThemeBase


logger = logging.getLogger('presalytics.story.revealer')


class Revealer(presalytics.story.components.Renderer):
    """
    Renders `presalytics.story.outline.StoryOutline` objects to 
    [reveal.js](https://github.com/hakimel/reveal.js/) presentations

    Parameters
    ----------
    story_outline : presalytics.story.outline.StoryOutline
        The presalytics StoryOutline to be rendered and presented
    
    Attributes
    -----------
    base : lxml.etree.Element
        An etree element containing the base html for each slide

    plugins : list of dict
        Plugin data that transform to html `<script>` and `<link>` tags through
        the rendering process

    plugin_mgr : presalytics.lib.plugins.base.PluginManager
        Sorts, validates, and renders plugins

    """
    base: lxml.etree.Element
    reveal_params: typing.Dict[str, typing.Any]

    __component_kind__ = 'revealer'

    def __init__(
            self,
            story_outline: 'StoryOutline',
            pages: typing.List[int] = None,
            **kwargs):
        super(Revealer, self).__init__(story_outline, **kwargs)
        logger.info("Initializing story render for {}".format(story_outline.title))
        self.story_outline.validate()
        if isinstance(pages, int):
            pages = [pages]
        elif not pages:
            pages = [p for p in range(0, len(self.story_outline.pages))]
        elif not isinstance(pages, list) or not isinstance(pages[0], int):
            raise presalytics.lib.exceptions.InvalidArgumentException(message='"pages" must be a list of integers')
        if len([p for p in pages if p >= len(self.story_outline.pages)]) > 0:
            raise presalytics.lib.exceptions.InvalidArgumentException(message='"pages" can only contain integers lower than the number of pages in the story')
        self.pages_to_render = pages
        self.base = self._make_base()
        logger.info("Loading plugins")
        reveal_params = {}
        for key, val in kwargs.items():
            if key in presalytics.lib.plugins.reveal.RevealConfigPlugin.default_config.keys():
                reveal_params.update({key: val})
        if len(self.pages_to_render) == 1:
            reveal_params.update({'controls': False})  # hide controls on single page story
        reveal_plugin_config = {
            'kind': 'script',
            'name': 'reveal',
            'config': {"reveal_params": reveal_params} if len(reveal_params.keys()) > 0 else {}
        }
        overrides_config = {
            'kind': 'style',
            'name': 'local',
            'config': {
                "css_file_id": "reveal_overrides"
            }
        }
        self.update_outline_from_instances()
        self.plugins = [reveal_plugin_config, overrides_config]
        self.get_component_implicit_plugins()
        outline_plugins = presalytics.lib.plugins.base.PluginManager.get_plugins_from_nested_dict(source_dict=self.story_outline.to_dict())
        self.plugins.extend(outline_plugins)
        self.plugin_mgr = presalytics.lib.plugins.base.PluginManager(self.plugins)
        logger.info("Revealer initilized.")

    def _make_base(self):
        base = lxml.etree.Element("div", attrib={
            "class": "reveal",
        })
        lxml.etree.SubElement(base, "div", attrib={"class": "slides"})

        try:
            story_id = self.story_outline.story_id
            base.attrib['data-story-id'] = story_id
        except Exception:
            logger.info("Revealer could not extract story_id from outline.")
        return base
    
    def get_meta_tags(self, body=tuple()):
        """
        Security Note: If supplying a body, ensure that its already been stripped of unauthorized scripts. 
        """

        scripts = body.findall(".//script")

        srcs = []
        for script in scripts:
            src = script.get("src")
            if src:
                root = urllib.parse.urlparse(src).netloc
                if root:
                    srcs.append(root)
        allowed = ' '.join(set(srcs))

        tags = [
            '<meta charset="utf-8">',
            '<meta http-equiv="X-UA-Compatible" content="IE=edge">',
            '<meta name="viewport" content="width=device-width, initial-scale=1, shrink-to-fit=no">',
        #    """<meta http-equiv="Content-Security-Policy" content="default-src 'self' https://*.presalytics.io; script-src 'self' https://*.presalytics.io {0};">""".format(allowed)
        ]
        return tags

    def package_as_standalone(self):
        """
        Render the story outline as a html document with only the 
        reveal.js presentation as conent

        Returns
        ----------
        A `str` containing a complete html document with the presentation
        """
        pres = self.render()
        body = E.BODY()
        body.append(pres)
        body = self.strip_unauthorized_scripts(body)
        for scripts in self.plugin_mgr.get_scripts():
            lxml_scripts = lxml.html.fragments_fromstring(scripts)
            for item in lxml_scripts:
                body.append(item)
        head = E.HEAD()
        for meta in self.get_meta_tags(body):
            lxml_meta = lxml.html.fragment_fromstring(meta)
            head.append(lxml_meta)
        for link in self.plugin_mgr.get_styles():
            lxml_links = lxml.html.fragments_fromstring(link)
            for item in lxml_links:
                head.append(item)
        head = self.strip_unauthorized_scripts(head)
        html = E.HTML(
            head,
            body
        )
        return lxml.html.tostring(html, pretty_print=True)

    # def render_plugin(self, plugin):
    #     plugin_type = plugin['type']
    #     plugin_config = plugin['config']
    #     plugin_class = self.plugin_loader.load(plugin_type)
    #     script_string = plugin_class().to_string(plugin_config)
    #     return lxml.html.fragment_fromstring(script_string)

    def update_info(self):
        """
        Updates story metadata
        """
        info = self.story_outline.info
        info.date_modified = datetime.datetime.utcnow()

    def render(self):
        """
        Creates a reveal.js presenation html fragement

        Returns:
        ---------
        A `str` html fragment containing a reveal.js presentation
        """
        reveal_base = self.base
        for p in range(0, len(self.story_outline.pages)):
            if p in self.pages_to_render:
                page = self.story_outline.pages[p]
                slides_container = reveal_base[0]
                slide = lxml.etree.SubElement(slides_container, "section")
                page_html = self.render_page(page)
                slide_fragment = lxml.html.fragment_fromstring(page_html)
                slide.append(slide_fragment)
        return reveal_base

    def render_page(self, page: 'Page') -> str:
        """
        Creates a reveal.js slide

        Returns
        ----------
        A `str` html framgment of the page
        """
        class_key = "page." + page.kind
        key = class_key + "." + page.name
        try:
            if presalytics.COMPONENTS.get_instance(key):
                page_instance = presalytics.COMPONENTS.get_instance(key)
            else:
                klass = presalytics.COMPONENTS.get(class_key)
                deserialize_method = getattr(klass, "deserialize", None)
                if callable(deserialize_method):
                    page_instance = deserialize_method(page, client_info=self.client_info)
                else:
                    message = 'Page component instance or class (kind) "{0}" unavailable in component registry'.format(key)
                    raise presalytics.lib.exceptions.MissingConfigException(message)
            page_html = page_instance.render()
        except Exception as ex:
            logger.exception(ex)
            t, v, tb = sys.exc_info()
            if not presalytics.CONFIG.get("DEBUG", False):
                page_html = presalytics.lib.exceptions.RenderExceptionHandler(ex, "page", traceback=tb).render_exception()
            else:
                six.reraise(t, v, tb)
        return page_html

    def present(self, files_path=None, debug=True, port=8082, host='127.0.0.1'):
        """
        Creates and opens the rendered story in the browser.  Story files are served by 
        a local flask server.  Not for production use.  Press Ctrl + C to close the server.

        Parameters
        ----------
        files_path : str
            filepath to a local folder that will work as root folder for a local flask
            server.  Defaults to the user's temporary files directory

        debug : str
            Defaults to True.  Indicates whether the flask server should be started
            in debug mode.
        
        port : str
            The network port to serve the story onto.  Defautls to 8082.
        
        host : str
            The host to for the local server.  Typically either localhost or the default gateway.
            Defaults to 127.0.0.1 (localhost).
        
        """
        logger.info("Building story rendering at http://{0}:{1}".format(host, port))
        if not files_path:
            files_path = tempfile.gettempdir()
        logger.info("Buidling standalone package for local rendering.")
        html = self.package_as_standalone().decode('utf-8')
        id = presalytics.story.util.to_title_case(self.story_outline.title)
        if id == '':
            id = 'blank'
        server = presalytics.story.server.LocalServer(host=host, debug=debug, port=port, root_path=files_path)
        pkg_templates_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), "lib", "templates")
        shutil.copy(os.path.join(pkg_templates_dir, "favicon.ico"), os.path.join(files_path, "presalytics", "static"))
        html_file = os.path.join(files_path, "presalytics", "templates", id + '.html')
        with open(html_file, 'w') as file:
            file.write(html)
        address = "http://{}:{}/story/{}".format(host, port, id)
        logger.info("Opening browser tab...")
        presalytics.story.server.Browser(address).start()
        server.run()



# coding: utf-8

"""
    OOXML Automation

    This API helps users convert Excel and Powerpoint documents into rich, live dashboards and stories.  # noqa: E501

    The version of the OpenAPI document: 0.1.0-no-tags
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six


class TableCellsDetails(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'row_id': 'str',
        'row': 'TableRowsDetails',
        'column_id': 'str',
        'column': 'TableColumnsDetails',
        'row_span': 'int',
        'column_span': 'int',
        'is_merged_vertical': 'bool',
        'is_merged_horozontal': 'bool',
        'fill_map': 'SharedFillMapDetails',
        'text_container': 'SharedTextContainerDetails',
        'border': 'TableBordersDetails',
        'id': 'str',
        'date_created': 'datetime',
        'user_created': 'str',
        'date_modified': 'datetime',
        'user_modified': 'str'
    }

    attribute_map = {
        'row_id': 'rowId',
        'row': 'row',
        'column_id': 'columnId',
        'column': 'column',
        'row_span': 'rowSpan',
        'column_span': 'columnSpan',
        'is_merged_vertical': 'isMergedVertical',
        'is_merged_horozontal': 'isMergedHorozontal',
        'fill_map': 'fillMap',
        'text_container': 'textContainer',
        'border': 'border',
        'id': 'id',
        'date_created': 'dateCreated',
        'user_created': 'userCreated',
        'date_modified': 'dateModified',
        'user_modified': 'userModified'
    }

    def __init__(self, row_id=None, row=None, column_id=None, column=None, row_span=None, column_span=None, is_merged_vertical=None, is_merged_horozontal=None, fill_map=None, text_container=None, border=None, id=None, date_created=None, user_created=None, date_modified=None, user_modified=None):  # noqa: E501
        """TableCellsDetails - a model defined in OpenAPI"""  # noqa: E501

        self._row_id = None
        self._row = None
        self._column_id = None
        self._column = None
        self._row_span = None
        self._column_span = None
        self._is_merged_vertical = None
        self._is_merged_horozontal = None
        self._fill_map = None
        self._text_container = None
        self._border = None
        self._id = None
        self._date_created = None
        self._user_created = None
        self._date_modified = None
        self._user_modified = None
        self.discriminator = None

        self.row_id = row_id
        if row is not None:
            self.row = row
        self.column_id = column_id
        if column is not None:
            self.column = column
        if row_span is not None:
            self.row_span = row_span
        if column_span is not None:
            self.column_span = column_span
        if is_merged_vertical is not None:
            self.is_merged_vertical = is_merged_vertical
        if is_merged_horozontal is not None:
            self.is_merged_horozontal = is_merged_horozontal
        if fill_map is not None:
            self.fill_map = fill_map
        if text_container is not None:
            self.text_container = text_container
        if border is not None:
            self.border = border
        if id is not None:
            self.id = id
        if date_created is not None:
            self.date_created = date_created
        if user_created is not None:
            self.user_created = user_created
        if date_modified is not None:
            self.date_modified = date_modified
        if user_modified is not None:
            self.user_modified = user_modified

    @property
    def row_id(self):
        """Gets the row_id of this TableCellsDetails.  # noqa: E501


        :return: The row_id of this TableCellsDetails.  # noqa: E501
        :rtype: str
        """
        return self._row_id

    @row_id.setter
    def row_id(self, row_id):
        """Sets the row_id of this TableCellsDetails.


        :param row_id: The row_id of this TableCellsDetails.  # noqa: E501
        :type: str
        """

        self._row_id = row_id

    @property
    def row(self):
        """Gets the row of this TableCellsDetails.  # noqa: E501


        :return: The row of this TableCellsDetails.  # noqa: E501
        :rtype: TableRowsDetails
        """
        return self._row

    @row.setter
    def row(self, row):
        """Sets the row of this TableCellsDetails.


        :param row: The row of this TableCellsDetails.  # noqa: E501
        :type: TableRowsDetails
        """

        self._row = row

    @property
    def column_id(self):
        """Gets the column_id of this TableCellsDetails.  # noqa: E501


        :return: The column_id of this TableCellsDetails.  # noqa: E501
        :rtype: str
        """
        return self._column_id

    @column_id.setter
    def column_id(self, column_id):
        """Sets the column_id of this TableCellsDetails.


        :param column_id: The column_id of this TableCellsDetails.  # noqa: E501
        :type: str
        """

        self._column_id = column_id

    @property
    def column(self):
        """Gets the column of this TableCellsDetails.  # noqa: E501


        :return: The column of this TableCellsDetails.  # noqa: E501
        :rtype: TableColumnsDetails
        """
        return self._column

    @column.setter
    def column(self, column):
        """Sets the column of this TableCellsDetails.


        :param column: The column of this TableCellsDetails.  # noqa: E501
        :type: TableColumnsDetails
        """

        self._column = column

    @property
    def row_span(self):
        """Gets the row_span of this TableCellsDetails.  # noqa: E501


        :return: The row_span of this TableCellsDetails.  # noqa: E501
        :rtype: int
        """
        return self._row_span

    @row_span.setter
    def row_span(self, row_span):
        """Sets the row_span of this TableCellsDetails.


        :param row_span: The row_span of this TableCellsDetails.  # noqa: E501
        :type: int
        """

        self._row_span = row_span

    @property
    def column_span(self):
        """Gets the column_span of this TableCellsDetails.  # noqa: E501


        :return: The column_span of this TableCellsDetails.  # noqa: E501
        :rtype: int
        """
        return self._column_span

    @column_span.setter
    def column_span(self, column_span):
        """Sets the column_span of this TableCellsDetails.


        :param column_span: The column_span of this TableCellsDetails.  # noqa: E501
        :type: int
        """

        self._column_span = column_span

    @property
    def is_merged_vertical(self):
        """Gets the is_merged_vertical of this TableCellsDetails.  # noqa: E501


        :return: The is_merged_vertical of this TableCellsDetails.  # noqa: E501
        :rtype: bool
        """
        return self._is_merged_vertical

    @is_merged_vertical.setter
    def is_merged_vertical(self, is_merged_vertical):
        """Sets the is_merged_vertical of this TableCellsDetails.


        :param is_merged_vertical: The is_merged_vertical of this TableCellsDetails.  # noqa: E501
        :type: bool
        """

        self._is_merged_vertical = is_merged_vertical

    @property
    def is_merged_horozontal(self):
        """Gets the is_merged_horozontal of this TableCellsDetails.  # noqa: E501


        :return: The is_merged_horozontal of this TableCellsDetails.  # noqa: E501
        :rtype: bool
        """
        return self._is_merged_horozontal

    @is_merged_horozontal.setter
    def is_merged_horozontal(self, is_merged_horozontal):
        """Sets the is_merged_horozontal of this TableCellsDetails.


        :param is_merged_horozontal: The is_merged_horozontal of this TableCellsDetails.  # noqa: E501
        :type: bool
        """

        self._is_merged_horozontal = is_merged_horozontal

    @property
    def fill_map(self):
        """Gets the fill_map of this TableCellsDetails.  # noqa: E501


        :return: The fill_map of this TableCellsDetails.  # noqa: E501
        :rtype: SharedFillMapDetails
        """
        return self._fill_map

    @fill_map.setter
    def fill_map(self, fill_map):
        """Sets the fill_map of this TableCellsDetails.


        :param fill_map: The fill_map of this TableCellsDetails.  # noqa: E501
        :type: SharedFillMapDetails
        """

        self._fill_map = fill_map

    @property
    def text_container(self):
        """Gets the text_container of this TableCellsDetails.  # noqa: E501


        :return: The text_container of this TableCellsDetails.  # noqa: E501
        :rtype: SharedTextContainerDetails
        """
        return self._text_container

    @text_container.setter
    def text_container(self, text_container):
        """Sets the text_container of this TableCellsDetails.


        :param text_container: The text_container of this TableCellsDetails.  # noqa: E501
        :type: SharedTextContainerDetails
        """

        self._text_container = text_container

    @property
    def border(self):
        """Gets the border of this TableCellsDetails.  # noqa: E501


        :return: The border of this TableCellsDetails.  # noqa: E501
        :rtype: TableBordersDetails
        """
        return self._border

    @border.setter
    def border(self, border):
        """Sets the border of this TableCellsDetails.


        :param border: The border of this TableCellsDetails.  # noqa: E501
        :type: TableBordersDetails
        """

        self._border = border

    @property
    def id(self):
        """Gets the id of this TableCellsDetails.  # noqa: E501


        :return: The id of this TableCellsDetails.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this TableCellsDetails.


        :param id: The id of this TableCellsDetails.  # noqa: E501
        :type: str
        """

        self._id = id

    @property
    def date_created(self):
        """Gets the date_created of this TableCellsDetails.  # noqa: E501


        :return: The date_created of this TableCellsDetails.  # noqa: E501
        :rtype: datetime
        """
        return self._date_created

    @date_created.setter
    def date_created(self, date_created):
        """Sets the date_created of this TableCellsDetails.


        :param date_created: The date_created of this TableCellsDetails.  # noqa: E501
        :type: datetime
        """

        self._date_created = date_created

    @property
    def user_created(self):
        """Gets the user_created of this TableCellsDetails.  # noqa: E501


        :return: The user_created of this TableCellsDetails.  # noqa: E501
        :rtype: str
        """
        return self._user_created

    @user_created.setter
    def user_created(self, user_created):
        """Sets the user_created of this TableCellsDetails.


        :param user_created: The user_created of this TableCellsDetails.  # noqa: E501
        :type: str
        """

        self._user_created = user_created

    @property
    def date_modified(self):
        """Gets the date_modified of this TableCellsDetails.  # noqa: E501


        :return: The date_modified of this TableCellsDetails.  # noqa: E501
        :rtype: datetime
        """
        return self._date_modified

    @date_modified.setter
    def date_modified(self, date_modified):
        """Sets the date_modified of this TableCellsDetails.


        :param date_modified: The date_modified of this TableCellsDetails.  # noqa: E501
        :type: datetime
        """

        self._date_modified = date_modified

    @property
    def user_modified(self):
        """Gets the user_modified of this TableCellsDetails.  # noqa: E501


        :return: The user_modified of this TableCellsDetails.  # noqa: E501
        :rtype: str
        """
        return self._user_modified

    @user_modified.setter
    def user_modified(self, user_modified):
        """Sets the user_modified of this TableCellsDetails.


        :param user_modified: The user_modified of this TableCellsDetails.  # noqa: E501
        :type: str
        """

        self._user_modified = user_modified

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, TableCellsDetails):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



import typing
import datetime
import presalytics.story.components
import presalytics.story.outline
if typing.TYPE_CHECKING:
    from presalytics.story.components import WidgetBase, PageTemplateBase
    from presalytics.story.outline import StoryOutline, Info, Page, Widget, Theme


def create_outline_from_widget(widget: 'WidgetBase',
                               page_name: str = None,
                               title: str = None,
                               description: str = None) -> 'StoryOutline':
    """
    Creates a `presalytics.story.outline.StoryOutline` from a subclass of 
    `presalytics.story.components.WidgetBase`.  Useful for quickstarts, demos, and 
    rapidly sharing content without having to manaully build a `StoryOutline`.

    Parameters
    ----------

    widget : subclass of presalytics.story.components.WidgetBase
        The widget component that you want to build the story from

    page_name : str, optional
        The name you want to give the `presalytics.story.outline.Page` in the outline.
        Defaults to the name of the supplied `widget`

    title : str, optional
        The title of the `presalytics.story.outline.StoryOutline`.  Defaults to
        `page_name`.

    description : str, optional
        The description of the story.  Autopopulated if unassigned.

    """
    info = presalytics.story.outline.Info(
        revision="0",
        date_created=datetime.datetime.now().astimezone(datetime.timezone.utc).isoformat(),
        date_modified=datetime.datetime.now().astimezone(datetime.timezone.utc).isoformat(),
        created_by=presalytics.CONFIG["USERNAME"],
        modified_by=presalytics.CONFIG["USERNAME"],
        revision_notes="Created by 'create_outline_from_widget' method"
    )

    outline_widget = widget.serialize()
    if not page_name:
        page_name = outline_widget.name

    page = presalytics.story.outline.Page(
        name=page_name,
        kind="widget-page",
        widgets=[outline_widget]
    )
    if not description:
        description = "{0} with {1} created by 'create_outline_from_widget' method".format(page_name, outline_widget.name)

    if not title:
        title = page_name

    outline = presalytics.story.outline.StoryOutline(
        info=info,
        pages=[page],
        description=description,
        title=title,
        themes=[]
    )

    return outline

def create_outline_from_page(page: 'PageTemplateBase',
                             title: str = None,
                             description: str = None) -> 'StoryOutline':
    """
    Creates a `presalytics.story.outline.StoryOutline` from a subclass of 
    `presalytics.story.components.PageTemplateBase`.  Useful for quick starts, demos, and 
    rapidly sharing content without having to manaully build a `StoryOutline`.

    Parameters
    ----------

    page : subclass of presalytics.story.components.PageTemplateBase
        The Page component that you want to build the story from

    title : str, optional
        The title of the `presalytics.story.outline.StoryOutline`.  Defaults to
        the name of the page.

    description : str, optional
        The description of the story.  Autopopulated if unassigned.

    """
    info = presalytics.story.outline.Info(
        revision="0",
        date_created=datetime.datetime.now().astimezone(datetime.timezone.utc).isoformat(),
        date_modified=datetime.datetime.now().astimezone(datetime.timezone.utc).isoformat(),
        created_by=presalytics.CONFIG["USERNAME"],
        modified_by=presalytics.CONFIG["USERNAME"],
        revision_notes="Created by 'create_outline_from_page' method"
    )

    page_outline = page.serialize()

    if not description:
        description = "{0} created by 'create_outline_from_page' method".format(page_outline.name)

    if not title:
        title = page_outline.name

    outline = presalytics.story.outline.StoryOutline(
        info=info,
        pages=[page],
        description=description,
        title=title,
        themes=[]
    )

    return outline


import json
import presalytics.lib.plugins.base as ext
import presalytics.lib.plugins.jinja as jinja

class Mpld3NumpyToJson(json.JSONEncoder):
    """
    Simplified Numpy encoder that returns json that mpld3 will understand 
    """
    def default(self, obj):
        import numpy as np
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


class Mpld3Plugin(ext.ScriptPlugin, jinja.JinjaPluginMakerMixin):
    __plugin_name__ = 'mpld3'

    __dependencies__ = [
        {
            'kind': 'script',
            'name': 'external_scripts',
            'config': {
                'approved_scripts_key': 'd3v3'
            }
        },
        {
            'kind': 'script',
            'name': 'external_scripts',
            'config': {
                'approved_scripts_key': 'mpld3'
            }
        }
    ]

    template = '<script type="text/javascript">mpld3.draw_figure("{{ id|safe }}",{{ fig_json }});</script>'

    def to_script(self, config, **kwargs):
        fig_json = json.dumps(config["figure_dict"], cls=Mpld3NumpyToJson)
        render_config = {
            'id': config["id"],
            'fig_json': fig_json
        }
        return self.render(render_config)




# coding: utf-8

"""
    Stroy API

    This api serves as and entry point for managing stories on the presaltytics api  # noqa: E501

    OpenAPI spec version: 3.0.2
    Generated by: https://openapi-generator.tech
"""


from setuptools import setup, find_packages  # noqa: H301

NAME = "presalytics_story"
VERSION = "0.1.63"

# To install the library, run the following
#
# python setup.py install
#
# prerequisite: setuptools
# http://pypi.python.org/pypi/setuptools

REQUIRES = [
    "urllib3 >= 1.15", 
    "six >= 1.10", 
    "certifi", 
    "python-dateutil"
]

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name=NAME,
    version=VERSION,
    description="Story API",
    author_email="kevin@presalytics.io",
    url="http://github.com/presalytics/story-python-client",
    keywords=["presalytics", "Story API"],
    install_requires=REQUIRES,
    packages=find_packages(),
    include_package_data=True,
    long_description_content_type="text/markdown",
    long_description=long_description
)



import unittest
import os
import json
import lxml
import lxml.html
import lxml.etree
import base64
import io
import uuid
import presalytics.story.revealer
import presalytics.story.outline
import presalytics.lib.tools.component_tools
import presalytics.lib.widgets.ooxml_editors
import presalytics.lib.widgets.d3


class TestStory(unittest.TestCase):
    """
    Test module features thatr render stories to dashboards or other formats.

    Please note that these are integration tests, the development environment and
    `presaltytics.CONFIG` must be set up appropriately for these test execute 
    sucessfully.
    """
    def setUp(self):
        pass

    def test_deserailze_yaml(self):
        test_file = os.path.join(os.path.dirname(__file__), 'files', 'matplotlib-outline.yaml')
        outline = presalytics.story.outline.StoryOutline.import_yaml(test_file)
        self.assertTrue(isinstance(outline, presalytics.story.outline.StoryOutline))
        self.assertTrue(isinstance(outline.info, presalytics.story.outline.Info))
        self.assertTrue(isinstance(outline.pages[0], presalytics.story.outline.Page))
        self.assertTrue(isinstance(outline.pages[0].widgets[0], presalytics.story.outline.Widget))
        self.assertTrue(isinstance(outline.themes[0], presalytics.story.outline.Theme))
        self.assertEqual('Test-story', outline.title)

    def test_serialize_to_json(self):
        test_file = os.path.join(os.path.dirname(__file__), 'files', 'matplotlib-outline.yaml')
        outline = presalytics.story.outline.StoryOutline.import_yaml(test_file)
        json_str = outline.dump()
        new_dict = json.loads(json_str)
        self.assertEqual(new_dict["pages"][0]["widgets"][0]["data"]["temp"], "data")
        self.assertIsNotNone(json_str)

    def test_plugins(self):
        from presalytics import PLUGINS

    def test_create_outline_from_widget(self):
        import matplotlib.pyplot as plt
        import numpy as np
        
        x = np.random.rand(30)
        y = np.random.rand(30)
        z = np.random.rand(30)
        
        fig, ax = plt.subplots()

        ax.scatter(x, y, s=z*1000, alpha=0.5)

        wrapper = presalytics.lib.widgets.matplotlib.MatplotlibResponsiveFigure(fig, "BubbleChart")


        outline = presalytics.lib.tools.component_tools.create_outline_from_widget(wrapper)

        re_wrapper = presalytics.lib.widgets.matplotlib.MatplotlibResponsiveFigure.deserialize(outline.pages[0].widgets[0])

    

        self.assertIsInstance(outline, presalytics.story.outline.StoryOutline)
        self.assertIsInstance(re_wrapper, presalytics.lib.widgets.matplotlib.MatplotlibFigure)

        revealer = presalytics.story.revealer.Revealer(outline)

        self.assertGreaterEqual(len(revealer.plugin_mgr.dependency_map), 8)

        html_str = revealer.package_as_standalone().decode('utf-8')

        self.assertIsInstance(html_str, str)

    def test_render_page_exception(self):
        test_file = os.path.join(os.path.dirname(__file__), 'files', 'bad-outline.yaml')
        _debug = presalytics.CONFIG.pop("DEBUG", None)
        outline = presalytics.story.outline.StoryOutline.import_yaml(test_file)
        revealer = presalytics.story.revealer.Revealer(outline)
        html = revealer.package_as_standalone().decode('utf-8')
        self.assertTrue("Oops!" in html)
        if _debug:
            presalytics.CONFIG.update({"DEBUG": _debug})
    
    def text_replace_transform_test(self):
        test_file = os.path.join(os.path.dirname(__file__), 'files', 'ooxml_test_2.xml')
        test_element = lxml.etree.parse(test_file)
        params = {
            'beta': '66.02',
            'fit_quality': 'pretty good',
            'r_squared': '91.33%',
            'trend': 'positive'
        }

        replacer = presalytics.lib.widgets.ooxml_editors.TextReplace(params)

        updates = lxml.etree.tostring(replacer.execute(test_element)).decode('utf-8')

        for _, val in params.items():
            self.assertTrue(val in updates)

    def upload_from_json(self):
        from presalytics.client.presalytics_story import FileUpload

        test_file = os.path.join(os.path.dirname(__file__), 'files', 'bubblechart.pptx')
        with open(test_file, 'rb') as f:
            bin = f.read()
            content_length = len(bin)
            data = base64.b64encode(bin).decode('UTF-8')
            
        client = presalytics.Client()
        file = FileUpload(
            content_length=content_length,
            file_name="test.pptx",
            mimetype='application/vnd.openxmlformats-officedocument.presentationml.presentation',
            file=str(data)
        )
        story = client.story.story_post_file_json(file_upload=file)
        from presalytics.client.presalytics_story import Story
        self.assertTrue(type(story) is Story)

    def d3(self):
        
        widget = presalytics.lib.widgets.d3.D3Widget(
            'test-widget',
            {
                'right_text': 'This is the right window',
                'left_text': 'This is the left window'
            },
            script_filename='d3-test.js',
            css_filename='fragment.css',
            html_filename='fragment.html'
        )
        widget.story_id = str(uuid.uuid4())
        w = widget.serialize()
        self.assertTrue(isinstance(w, presalytics.story.outline.Widget))
        new_widget = presalytics.lib.widgets.d3.D3Widget.deserialize(w)
        self.assertTrue(isinstance(new_widget, presalytics.lib.widgets.d3.D3Widget))

        html = widget.standalone_html()

        doc = lxml.html.document_fromstring(html)

        self.assertTrue(isinstance(doc, lxml.etree._Element))

        frame = lxml.html.fragment_fromstring(widget.to_html())

        self.assertTrue(isinstance(frame, lxml.etree._Element))

        # Uncomment to view test results

        # client = presalytics.Client()

        # outline = presalytics.lib.tools.component_tools.create_outline_from_widget(widget)

        # client.story.story_post({"outline": outline.dump()})ks
        


    def tearDown(self):
        pass



# coding: utf-8

"""
    OOXML Automation

    This API helps users convert Excel and Powerpoint documents into rich, live dashboards and stories.  # noqa: E501

    The version of the OpenAPI document: 0.1.0-no-tags
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six


class SlideGroupElementTypesDetails(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'type_id': 'int',
        'name': 'str',
        'description': 'str',
        'id': 'str',
        'date_created': 'datetime',
        'user_created': 'str',
        'date_modified': 'datetime',
        'user_modified': 'str'
    }

    attribute_map = {
        'type_id': 'typeId',
        'name': 'name',
        'description': 'description',
        'id': 'id',
        'date_created': 'dateCreated',
        'user_created': 'userCreated',
        'date_modified': 'dateModified',
        'user_modified': 'userModified'
    }

    def __init__(self, type_id=None, name=None, description=None, id=None, date_created=None, user_created=None, date_modified=None, user_modified=None):  # noqa: E501
        """SlideGroupElementTypesDetails - a model defined in OpenAPI"""  # noqa: E501

        self._type_id = None
        self._name = None
        self._description = None
        self._id = None
        self._date_created = None
        self._user_created = None
        self._date_modified = None
        self._user_modified = None
        self.discriminator = None

        if type_id is not None:
            self.type_id = type_id
        self.name = name
        self.description = description
        if id is not None:
            self.id = id
        if date_created is not None:
            self.date_created = date_created
        if user_created is not None:
            self.user_created = user_created
        if date_modified is not None:
            self.date_modified = date_modified
        if user_modified is not None:
            self.user_modified = user_modified

    @property
    def type_id(self):
        """Gets the type_id of this SlideGroupElementTypesDetails.  # noqa: E501


        :return: The type_id of this SlideGroupElementTypesDetails.  # noqa: E501
        :rtype: int
        """
        return self._type_id

    @type_id.setter
    def type_id(self, type_id):
        """Sets the type_id of this SlideGroupElementTypesDetails.


        :param type_id: The type_id of this SlideGroupElementTypesDetails.  # noqa: E501
        :type: int
        """

        self._type_id = type_id

    @property
    def name(self):
        """Gets the name of this SlideGroupElementTypesDetails.  # noqa: E501


        :return: The name of this SlideGroupElementTypesDetails.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this SlideGroupElementTypesDetails.


        :param name: The name of this SlideGroupElementTypesDetails.  # noqa: E501
        :type: str
        """

        self._name = name

    @property
    def description(self):
        """Gets the description of this SlideGroupElementTypesDetails.  # noqa: E501


        :return: The description of this SlideGroupElementTypesDetails.  # noqa: E501
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """Sets the description of this SlideGroupElementTypesDetails.


        :param description: The description of this SlideGroupElementTypesDetails.  # noqa: E501
        :type: str
        """

        self._description = description

    @property
    def id(self):
        """Gets the id of this SlideGroupElementTypesDetails.  # noqa: E501


        :return: The id of this SlideGroupElementTypesDetails.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this SlideGroupElementTypesDetails.


        :param id: The id of this SlideGroupElementTypesDetails.  # noqa: E501
        :type: str
        """

        self._id = id

    @property
    def date_created(self):
        """Gets the date_created of this SlideGroupElementTypesDetails.  # noqa: E501


        :return: The date_created of this SlideGroupElementTypesDetails.  # noqa: E501
        :rtype: datetime
        """
        return self._date_created

    @date_created.setter
    def date_created(self, date_created):
        """Sets the date_created of this SlideGroupElementTypesDetails.


        :param date_created: The date_created of this SlideGroupElementTypesDetails.  # noqa: E501
        :type: datetime
        """

        self._date_created = date_created

    @property
    def user_created(self):
        """Gets the user_created of this SlideGroupElementTypesDetails.  # noqa: E501


        :return: The user_created of this SlideGroupElementTypesDetails.  # noqa: E501
        :rtype: str
        """
        return self._user_created

    @user_created.setter
    def user_created(self, user_created):
        """Sets the user_created of this SlideGroupElementTypesDetails.


        :param user_created: The user_created of this SlideGroupElementTypesDetails.  # noqa: E501
        :type: str
        """

        self._user_created = user_created

    @property
    def date_modified(self):
        """Gets the date_modified of this SlideGroupElementTypesDetails.  # noqa: E501


        :return: The date_modified of this SlideGroupElementTypesDetails.  # noqa: E501
        :rtype: datetime
        """
        return self._date_modified

    @date_modified.setter
    def date_modified(self, date_modified):
        """Sets the date_modified of this SlideGroupElementTypesDetails.


        :param date_modified: The date_modified of this SlideGroupElementTypesDetails.  # noqa: E501
        :type: datetime
        """

        self._date_modified = date_modified

    @property
    def user_modified(self):
        """Gets the user_modified of this SlideGroupElementTypesDetails.  # noqa: E501


        :return: The user_modified of this SlideGroupElementTypesDetails.  # noqa: E501
        :rtype: str
        """
        return self._user_modified

    @user_modified.setter
    def user_modified(self, user_modified):
        """Sets the user_modified of this SlideGroupElementTypesDetails.


        :param user_modified: The user_modified of this SlideGroupElementTypesDetails.  # noqa: E501
        :type: str
        """

        self._user_modified = user_modified

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, SlideGroupElementTypesDetails):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



# coding: utf-8

"""
    OOXML Automation

    This API helps users convert Excel and Powerpoint documents into rich, live dashboards and stories.  # noqa: E501

    The version of the OpenAPI document: 0.1.0-no-tags
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six


class SharedPictures(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'graphics_id': 'str',
        'image_fills_id': 'str',
        'image_file_blob_url': 'str',
        'file_extension': 'str',
        'base_element_blob_url': 'str',
        'changed_base_element_blob_url': 'str',
        'package_uri': 'str',
        'name': 'str',
        'id': 'str'
    }

    attribute_map = {
        'graphics_id': 'graphicsId',
        'image_fills_id': 'imageFillsId',
        'image_file_blob_url': 'imageFileBlobUrl',
        'file_extension': 'fileExtension',
        'base_element_blob_url': 'baseElementBlobUrl',
        'changed_base_element_blob_url': 'changedBaseElementBlobUrl',
        'package_uri': 'packageUri',
        'name': 'name',
        'id': 'id'
    }

    def __init__(self, graphics_id=None, image_fills_id=None, image_file_blob_url=None, file_extension=None, base_element_blob_url=None, changed_base_element_blob_url=None, package_uri=None, name=None, id=None):  # noqa: E501
        """SharedPictures - a model defined in OpenAPI"""  # noqa: E501

        self._graphics_id = None
        self._image_fills_id = None
        self._image_file_blob_url = None
        self._file_extension = None
        self._base_element_blob_url = None
        self._changed_base_element_blob_url = None
        self._package_uri = None
        self._name = None
        self._id = None
        self.discriminator = None

        self.graphics_id = graphics_id
        self.image_fills_id = image_fills_id
        self.image_file_blob_url = image_file_blob_url
        self.file_extension = file_extension
        self.base_element_blob_url = base_element_blob_url
        self.changed_base_element_blob_url = changed_base_element_blob_url
        self.package_uri = package_uri
        self.name = name
        if id is not None:
            self.id = id

    @property
    def graphics_id(self):
        """Gets the graphics_id of this SharedPictures.  # noqa: E501


        :return: The graphics_id of this SharedPictures.  # noqa: E501
        :rtype: str
        """
        return self._graphics_id

    @graphics_id.setter
    def graphics_id(self, graphics_id):
        """Sets the graphics_id of this SharedPictures.


        :param graphics_id: The graphics_id of this SharedPictures.  # noqa: E501
        :type: str
        """

        self._graphics_id = graphics_id

    @property
    def image_fills_id(self):
        """Gets the image_fills_id of this SharedPictures.  # noqa: E501


        :return: The image_fills_id of this SharedPictures.  # noqa: E501
        :rtype: str
        """
        return self._image_fills_id

    @image_fills_id.setter
    def image_fills_id(self, image_fills_id):
        """Sets the image_fills_id of this SharedPictures.


        :param image_fills_id: The image_fills_id of this SharedPictures.  # noqa: E501
        :type: str
        """

        self._image_fills_id = image_fills_id

    @property
    def image_file_blob_url(self):
        """Gets the image_file_blob_url of this SharedPictures.  # noqa: E501


        :return: The image_file_blob_url of this SharedPictures.  # noqa: E501
        :rtype: str
        """
        return self._image_file_blob_url

    @image_file_blob_url.setter
    def image_file_blob_url(self, image_file_blob_url):
        """Sets the image_file_blob_url of this SharedPictures.


        :param image_file_blob_url: The image_file_blob_url of this SharedPictures.  # noqa: E501
        :type: str
        """

        self._image_file_blob_url = image_file_blob_url

    @property
    def file_extension(self):
        """Gets the file_extension of this SharedPictures.  # noqa: E501


        :return: The file_extension of this SharedPictures.  # noqa: E501
        :rtype: str
        """
        return self._file_extension

    @file_extension.setter
    def file_extension(self, file_extension):
        """Sets the file_extension of this SharedPictures.


        :param file_extension: The file_extension of this SharedPictures.  # noqa: E501
        :type: str
        """

        self._file_extension = file_extension

    @property
    def base_element_blob_url(self):
        """Gets the base_element_blob_url of this SharedPictures.  # noqa: E501


        :return: The base_element_blob_url of this SharedPictures.  # noqa: E501
        :rtype: str
        """
        return self._base_element_blob_url

    @base_element_blob_url.setter
    def base_element_blob_url(self, base_element_blob_url):
        """Sets the base_element_blob_url of this SharedPictures.


        :param base_element_blob_url: The base_element_blob_url of this SharedPictures.  # noqa: E501
        :type: str
        """

        self._base_element_blob_url = base_element_blob_url

    @property
    def changed_base_element_blob_url(self):
        """Gets the changed_base_element_blob_url of this SharedPictures.  # noqa: E501


        :return: The changed_base_element_blob_url of this SharedPictures.  # noqa: E501
        :rtype: str
        """
        return self._changed_base_element_blob_url

    @changed_base_element_blob_url.setter
    def changed_base_element_blob_url(self, changed_base_element_blob_url):
        """Sets the changed_base_element_blob_url of this SharedPictures.


        :param changed_base_element_blob_url: The changed_base_element_blob_url of this SharedPictures.  # noqa: E501
        :type: str
        """

        self._changed_base_element_blob_url = changed_base_element_blob_url

    @property
    def package_uri(self):
        """Gets the package_uri of this SharedPictures.  # noqa: E501


        :return: The package_uri of this SharedPictures.  # noqa: E501
        :rtype: str
        """
        return self._package_uri

    @package_uri.setter
    def package_uri(self, package_uri):
        """Sets the package_uri of this SharedPictures.


        :param package_uri: The package_uri of this SharedPictures.  # noqa: E501
        :type: str
        """

        self._package_uri = package_uri

    @property
    def name(self):
        """Gets the name of this SharedPictures.  # noqa: E501


        :return: The name of this SharedPictures.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this SharedPictures.


        :param name: The name of this SharedPictures.  # noqa: E501
        :type: str
        """

        self._name = name

    @property
    def id(self):
        """Gets the id of this SharedPictures.  # noqa: E501


        :return: The id of this SharedPictures.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this SharedPictures.


        :param id: The id of this SharedPictures.  # noqa: E501
        :type: str
        """

        self._id = id

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, SharedPictures):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



import abc
import typing
import os
import logging
import webbrowser
import re
import six
import sys
import urllib.parse
import presalytics.lib
import presalytics.lib.registry
import presalytics.lib.exceptions
import presalytics.lib.constants
import presalytics.client.api
if typing.TYPE_CHECKING:
    from presalytics.story.outline import Widget, Page, Plugin, OutlineBase, StoryOutline


logger = logging.getLogger("presalytics.story.components")


class ComponentBase(abc.ABC):
    """
    This class serves as the primary interface clas for the story components.  Instances of
    this interface carry enough information to rendered to html.  Developers, analysts and
    users should inherit from base class if they want components to be registered in the
    rendering pipeline for presalytics stories.

    Instance of this class must:
    1. Initialize with a name parameter
    2  Serialize to a subclass of presalytics.story.outline.OutlineBase
    3. Deserialize from an instance of presalytics.story.OutlineBase
    4. Render to html

    The __component_type__ and __component_name__ metadata on subclasses is required in
    order for instances to be registered in rendering pipeline for automatic updates

    Note: When inheriting from this base class, call "super().__init__(self, *args, **kwargs)" before add your
    own custom initialization.

    Attributes
    ----------

    name : str
        The name of the instance of the component.  Used in component as a key to lookup
        instances to re-calcute during rendering

    css : list of str
        A list of keys in dot notation that point to links in the ALLOWED_STYLES dictionary
        in the presalytics.lib.templates.base module.

    js: list of str
        A list of keys in dot notation that point to links in the ALLOWED_SCRIPTS dictionary
        in the presalytics.lib.templates.base module.

    __component_type__ : str
        The identifier for the component superclass (e.g. widget, page_template, renderer, theme).
        Used for component registration.

    __component_kind__ : str
        An identifier for this component class.  Used for component registration.

    __plugins__ : list of dict
        A list of dictionaries that reference `presalytics.story.outline.Plugin` configurations.  When a 
        `presaltytics.story.components.Renderer` is initialized, it will load these plugins into the
        rendered.  This allows plugins to be statically configured on `presalytics.story.components` classes,
        in lieu dynamic configurations on `presalytics.story.outline.StoryOutline` instances.

    client_info : str
        A `dict` to be unpacked and passed to `presalytics.client.api.Client` object when subclasses of
        `ComponentBase` require interaction with the Presalytics API.
    """
    __component_type__: str
    __component_kind__: str
    __plugins__: typing.List[typing.Optional[typing.Dict[str, typing.Any]]]
    name: str

    __plugins__ = []

    def __init__(self, client_info: typing.Dict = None, *args, **kwargs):
        if client_info:
            self.client_info = client_info
        else:
            self.client_info = {}
            if kwargs.get('delegate_login', None):
                self.client_info.update({'delegate_login': kwargs.get('delegate_login')})
            if kwargs.get('cache_tokens', None):
                self.client_info.update({'cache_tokens': kwargs.get('cache_tokens')})
            if kwargs.get('token', None):
                self.client_info.update({'token': kwargs.get('token')})

    @abc.abstractmethod
    def render(self, **kwargs):
        """
        Renders `component` to html.  Must be overridden in subclass.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def serialize(self):
        """
        Initializes `component` object from `presalytics.story.outline` attributes. Must be overridden in subclass.
        """
        raise NotImplementedError

    @classmethod
    def deserialize(cls, component, **kwargs):
        """
        Converts `component` instance to a `presalytics.story.outline` object. Must be overridden in subclass.
        """
        raise NotImplementedError

    def get_client(self, **kwargs):
        """
        Initializes `presalytics.client.api.Client` using `client_info`
        """
        params = self.client_info
        params.update(kwargs)
        return presalytics.client.api.Client(**params)


class WidgetBase(ComponentBase):
    """
    Inherit from this base class to create widget components that can be rendered to html via the
    `presalytics.story.components.Renderer` class.  This component also needs to build a method
    that allows the widget to be serialzed into a `presalytics.story.outline.Widget` object.

    Parameters
    ----------
    widget: Widget
        A `presalytics.story.outline.Widget` object use for initialized the component class.

    Attributes
    ----------
    outline_widget: Widget
        A `presalytics.story.outline.Widget` object
    """
    outline_widget: typing.Optional['Widget']

    __component_type__ = 'widget'

    def __init__(self, name, *args, **kwargs) -> None:
        super(WidgetBase, self).__init__(*args, **kwargs)
        self.name = name
        self.outline_widget = None

    def render(self, component, **kwargs):
        self.to_html(component, **kwargs)

    @abc.abstractmethod
    def to_html(self, data: typing.Dict = None, **kwargs) -> str:
        """
        Returns valid html that renders the widget in a browser.

        Parameters
        ----------
        data: dict
            The data parameter is a dictionary should contain the minimum amount of that is required to
            successfully render the object.  As the widget is update, data control how the display of
            information changes.

        **kwargs:
            Optional keyword arguments can be used in subclass to modify the behavior of the `to_html` function.
            these keyword arguments should be invariant through successive updates to the chart.  For example,
            keyword arguments could control the styling of the widget, which should not change as the data in
            the object (e.g., a chart) is updated.  Keyword arguments are loaded via `additional_properties`
            parameter in in the `presalytics.story.outline.Widget` object.

        Returns
        ----------
        A str of containing an html fragment that will be loaded into a template in successive operations
        """
        raise NotImplementedError

    @abc.abstractmethod
    def serialize(self, **kwargs) -> 'Widget':
        """
        Creates `presalytics.story.outline.Widget` object from instance data. This widget should
        have the correct `name`, `data` and `additional_properties` so the same widget can be reconstituted
        via the to_html method, given the same set of data.

        Typically, this method will call an update method that run a local script with updates this
        Widget's data Dictionary prior being loading into the Widget outline object for serialization.

        Parameters
        ----------
        **kwargs:
            Optional keyword arguments can be used in subclass to modify the behavior of the to_html function.
            these keyword arguments should be be invariant through successive updates to the chart. Overrides
            for this widgets default additional_properties should be loaded via these keyword arguments.

        """
        raise NotImplementedError

    @classmethod
    def deserialize(cls, component: 'Widget', **kwargs) -> 'WidgetBase':
        """
        Creates an instance of the widget from the data object in the `presalytics.story.outline.Widget`
        object. This method exists to ensure widgets can be portable across environments.  To clarify,
        widgets built on the client-side via the `__init__` method can be reconstructed server-side via
        the deserialize method.  This allows decoupling of the widget generation/updating of data and
        the rendering of the widget in a UI.  Renderers (e.g., `presalytics.story.revealer.Revealer` object)
        need not know about how the data get updated, but can update the graphic with data generated by
        the widget when the serialize method is called.

        Parameters
        ----------

        widget: Widget
            A `presalytics.story.outline.Widget` object

        Returns
        ----------
        `presalytics.story.components.WidgetBase` subclass instance
        """
        raise NotImplementedError


class PageTemplateBase(ComponentBase):
    """
    Inherit from this base class to render templates to html via the
    `presalytics.story.revealer.Revealer` class.

    Parameters
    ----------
    page: Page
        A presalytics.story.outline.Page object for instalizing the class

    Attributes
    ----------
    outline_page: presalytics.story.outline.Page
        The page data

    widgets: list of subclasses of presalytics.story.components.WidgetBase
        A list widget that will be loaded into templates and rendered via placeholders.
        These widgets must have a "to_html(self, data, **kwargs)" method.

    """

    outline_page: 'Page'
    widgets: typing.List['WidgetBase']
    plugins: typing.List[typing.Dict]

    __component_type__ = 'page'

    def __init__(self, page: 'Page', **kwargs) -> None:
        super(PageTemplateBase, self).__init__(**kwargs)
        self.outline_page = page
        self.widgets = self.get_page_widgets(self.outline_page)
        

    @abc.abstractmethod
    def render(self, **kwargs) -> str:
        """
        Returns valid html that renders the template in a broswer with data loaded from widgets.
        """
        raise NotImplementedError

    def load_widget(self, widget: 'Widget'):
        """
        Converts a presalytics.story.outline.Widget object to a subclass of WidgetComponentBase
        via the `presalytics.COMPONENTS` registry.
        """
        class_key = "widget." + widget.kind
        key = class_key + "." + widget.name
        try:
            if presalytics.COMPONENTS.get_instance(key):
                widget_instance = presalytics.COMPONENTS.get_instance(key)
            else:
                klass = presalytics.COMPONENTS.get(class_key)
                deserialize_method = getattr(klass, "deserialize", None)
                if callable(deserialize_method):
                    widget_instance = deserialize_method(widget, client_info=self.client_info)
                else:
                    message = "Widget component instance or class (kind) {0} unavailable in component registry".format(key)
                    raise presalytics.lib.exceptions.MissingConfigException(message)
        except Exception as ex:
            logger.exception(ex)
            if not presalytics.CONFIG.get("DEBUG", False):
                widget_instance = presalytics.lib.exceptions.RenderExceptionHandler(ex)
            else:
                t, v, tb = sys.exc_info()
                six.reraise(t, v, tb)
        return widget_instance

    def get_page_widgets(self, page: 'Page'):
        """
        Converts the widgets within a `presalytics.story.outline.Page` object to a list
        of widgets subclassed from `presalytics.story.components.WidgetBase`
        """
        widget_instances = []
        for widget_outline in page.widgets:
            next_widget = self.load_widget(widget_outline)
            widget_instances.append(next_widget)
        return widget_instances
    
    def serialize(self):
        return self.outline_page.dump()

    @classmethod
    def deseriailize(cls, component, **kwargs):
        return cls(page=component)


class ThemeBase(ComponentBase):
    """
    Themes are containers for plugins should be rendered once
    across the entire document.  The init method should configure
    parameters with get passed to plugins via serialization and 
    deserialization.
    """
    plugins: typing.Optional[typing.List['Plugin']]

    __component_type__ = 'theme'

    def __init__(self, **kwargs):
        pass

    def render(self, **kwargs):
        pass


class Renderer(ComponentBase):
    """
    Base class for objects that convert `presalytics.story.outline.StoryOutline` 
    objects into html and rendering them over the web

    With this class, users can push changes to their `presalytics.story.outline.StoryOutline`
    to the Presalytics API and web clients.  Renderer class contains a couplemethods for 
    syncing changes from component instances in the `presalytics.CONFIG` to the Presalytics API 
    Story service.

    * The `view` method allows users programattically view their stories at https://presalytics.io 
    after changes are made

    * The `manage` method takes users to to the story management interface, where users can share their 
    work with other users, continue making edits or change story properties.


    Parameters
    ----------
    story_outline : presalytics.story.outline.StoryOutline
        The presalytics StoryOutline to be rendered and presented
    
    Attributes
    -----------
    plugins : list of dict
        Plugin data that transform to html `<script>` and `<link>` tags through
        the rendering process
    
    site_host : str
        The host of the website.  Defaults to https://presalytics.io.

    view_url : str, optional
         The url to view the story.  Unavailable if story outline has not been pushed to
         the Presalytics API story service

    manage_url : str, optional
         The url to view the story.  Unavailable if story outline has not been pushed to
         the Presalytics API story service
    """
    story_outline: 'StoryOutline'
    plugins: typing.List[typing.Dict]
    view_url: typing.Optional[str]
    manage_url: typing.Optional[str]

    __component_type__ = 'renderer'
    
    def __init__(self, story_outline : 'StoryOutline', **kwargs):
        super(Renderer, self).__init__(**kwargs)
        self.story_outline = story_outline
        try:
            self.site_host = presalytics.CONFIG["HOSTS"]["SITE"]
        except (KeyError, AttributeError):
            self.site_host = presalytics.lib.constants.SITE_HOST
        try:
            story_id = self.story_outline.story_id
            view_endpoint = presalytics.lib.constants.STORY_VIEW_URL.format(story_id)
            self.view_url = urllib.parse.urljoin(self.site_host, view_endpoint)
            manage_endpoint = presalytics.lib.constants.STORY_MANAGE_URL.format(story_id)
            self.manage_url = urllib.parse.urljoin(self.site_host, manage_endpoint)
        except (KeyError, AttributeError):
            self.view_url = None
            self.manage_url = None

    def render(self, **kwargs):
        """
        Renders server-side (i.e., opens the story on https://presalytics.io)
        """
        return self.view()

            
    
    def strip_unauthorized_scripts(self, body):
        """
        Finds and removes unauthorized scripts from that the html document.  For security reasons,
        content in `<script>` tags that has not been vetted by presalytics.io devops 

        If you would like to get a tag included in the base library, raise an issue on 
        [Github](https://github.com/presalytics/python-client/issues/new).  We'd love to hear from you and learn 
        about your use case, and will respond promptly to help.
        """
        allowed_scripts = presalytics.lib.plugins.external.ApprovedExternalScripts().attr_dict.flatten().values()
        script_elements = body.findall(".//script")
        for ele in script_elements:
            remove_ele = True
            script_id = ele.get("id")
            if script_id in self.story_outline.allowed_ids:
                remove_ele = False
            link = ele.get("src")
            if link in allowed_scripts:
                remove_ele = False
            if remove_ele:
                ele.getparent().remove(ele)
        return body
        
    @classmethod
    def deserialize(cls, component: 'StoryOutline', **kwargs):
        """
        Initializes the class from a `presalytics.story.outline.StoryOutline`.  See __init___.
        """
        return cls(component, **kwargs)

    def serialize(self) -> 'StoryOutline':
        """
        Updates the story_outline

        Returns
        -----------
        A refreshed `presalytics.story.outline.StoryOutline`
        """
        self.update_outline_from_instances()
        return self.story_outline

    def update_outline_from_instances(self, sub_dict: typing.Dict = None):
        """
        If a component instance for the widget is available in `presalytics.COMPONENTS`, 
        this method find the instance and regenerates the component data 
        so the latest data is available during the renering process.
        """
        if not sub_dict:
            sub_dict = self.story_outline.to_dict()
        if sub_dict:
            for key, val in sub_dict.items():
                if key in ["widgets", "themes", "pages"]:
                    if isinstance(val, list):
                        for list_item in val:
                            if isinstance(list_item, dict):
                                if "kind" in list_item:
                                    class_key = key.rstrip("s") + "." + list_item["kind"]
                                    klass = presalytics.COMPONENTS.get(class_key)
                                    if klass:
                                        if "name" in list_item:
                                            instance_key = class_key + "." + list_item["name"]
                                            inst = presalytics.COMPONENTS.get_instance(instance_key)
                                            if inst:
                                                self._set_outline_data_from_instance(inst)
                if isinstance(val, dict):
                    if len(val.keys()) > 0:
                        self.update_outline_from_instances(val)
                if isinstance(val, list):
                    for list_item in val:
                        if isinstance(list_item, dict):
                            self.update_outline_from_instances(list_item)

    def get_component_implicit_plugins(self, sub_dict: typing.Dict = None):
        """
        Retrieves plugin data from plugins attached to `presalytics.story.components`
        classes referenced in the `presalytics.story.outline.StoryOutline`
        """
        if not sub_dict:
            sub_dict = self.story_outline.to_dict()
        if sub_dict:
            for key, val in sub_dict.items():
                if key in ["widgets", "themes", "pages"]:
                    if isinstance(val, list):
                        for list_item in val:
                            if isinstance(list_item, dict):
                                if "kind" in list_item:
                                    class_key = key.rstrip("s") + "." + list_item["kind"]
                                    klass = presalytics.COMPONENTS.get(class_key)
                                    if klass:
                                        if len(klass.__plugins__) > 0:
                                            self.plugins.extend(klass.__plugins__)                           
                if isinstance(val, dict):
                    if len(val.keys()) > 0:
                        self.get_component_implicit_plugins(val)
                if isinstance(val, list):
                    for list_item in val:
                        if isinstance(list_item, dict):
                            self.get_component_implicit_plugins(list_item)

    def _set_outline_data_from_instance(self, inst):
        
        if inst.__component_type__ == 'widget':
            self._set_widget_outline_data(inst)
        if inst.__component_type__ == 'page':
            self._set_page_outline_data(inst)
        if inst.__component_type__ == 'theme':
            self._set_theme_outline_data(inst)

    def _set_theme_outline_data(self, inst: 'ThemeBase'):
        theme_index = None
        for t in range(0, len(self.story_outline.themes)):
            if inst.name == self.story_outline.themes[t].name:
                theme_index = t
            if theme_index:
                break
        theme_outline = inst.serialize()
        if theme_index:
            self.story_outline.themes[theme_index] = theme_outline    

    def _set_page_outline_data(self, inst: 'PageTemplateBase'):
        page_index = None
        for p in range(0, len(self.story_outline.pages)):
            if inst.name == self.story_outline.pages[p].name:
                page_index = p
            if page_index:
                break
        page_outline = inst.serialize()
        if page_index:
            self.story_outline.pages[page_index] = page_outline

    def _set_widget_outline_data(self, inst: 'WidgetBase'):
        widget_index: typing.Optional[int]
        page_index: typing.Optional[int]
        widget_index = None
        page_index = None
        for p in range(0, len(self.story_outline.pages)):
            for w in range(0, len(self.story_outline.pages[p].widgets)):
                widget = self.story_outline.pages[p].widgets[w]
                if widget.name == inst.name:
                    page_index = p
                    widget_index = w
                if page_index:
                    break
            if page_index:
                break
        w_outline = inst.serialize()
        if isinstance(page_index, int) and isinstance(widget_index, int): #  Causes 'unsupported target for assingment error`
            self.story_outline.pages[page_index].widgets[widget_index] = w_outline #type: ignore

    def update_story(self):
        """
        Updates the StoryOutline and pushes those updates to the Presalytics API Story service
        """
        self.update_outline_from_instances()
        client = presalytics.client.api.Client(**self.client_info)
        story = client.story.story_id_get(self.story_outline.story_id)
        story.outline = self.story_outline.dump()
        client.story.story_id_put(story.id, story)
    
    def view(self, update=False):
        """
        Updates a story and opens it on the presalytics.io website

        Parameters
        ----------
        update : bool
            Defaults to True.  Indicates whether the StoryOutline should be updated
            prior to opening in the web browser
        """
        if not self.view_url:
            message = "The outline has not been pushed to the Presalytics API yet, and therefore cannot be viewed via preslaytics.io"
            raise presalytics.lib.exceptions.InvalidConfigurationError(message=message)
        if update:
            self.update_story()
        webbrowser.open_new_tab(self.view_url)


    def manage(self, update=False):
        """
        Updates a story and opens the management page on the presalytics.io website

        Parameters
        ----------
        update : bool
            Defaults to True.  Indicates whether the StoryOutline should be updated
            prior to opening in the web browser
        """
        if not self.manage_url:
            message = "The outline has not been pushed to the Presalytics API yet, and therefore cannot be viewed via preslaytics.io"
            raise presalytics.lib.exceptions.InvalidConfigurationError(message=message)
            
        if update:
            self.update_story()
        webbrowser.open_new_tab(self.manage_url)


class ComponentRegistry(presalytics.lib.registry.RegistryBase):
    """
    A registry of classes and class instances of objects that inherit from
    `presalytics.story.components.ComponentBase`.

    """
    def __init__(self, **kwargs):
        self.instances = {}
        self.instance_regex = re.compile(r'(.*)\.(.*)\.(.*)')
        super(ComponentRegistry, self).__init__(**kwargs)

    def get_type(self, klass):
        """
        Returns the `__component_type__` attribute on a class
        """
        return getattr(klass, "__component_type__", None)

    def get_name(self, klass):
        """
        Returns the `__component_kind__` attribute on a class
        """
        return getattr(klass, "__component_kind__", None)

    def get_instance_name(self, klass):
        """
        Returns the `name` attribute on an instance
        """
        return getattr(klass, "name", None)

    def get_instance_registry_key(self, klass):
        """
        Creates a registry key from a class instance by concatenating the 
        `__component_type__`, `__component_kind__`, and `name` attributes of an instance
        """
        key = None
        klass_type = self.get_type(klass)
        if klass_type:
            klass_name = self.get_name(klass)
            instance_name = self.get_instance_name(klass)
            if instance_name and klass_name:
                key = "{0}.{1}.{2}".format(klass_type, klass_name, instance_name)
            else:
                # For now. always show this error, good for user debugging, rather than developer debugging
                message = '{0} instance missing "__component_kind__" or "name" attribute'.format(klass_type)
                logger.error(message)
        return key


    def load_class(self, klass):
        """
        Loads a class or instance into the registry
        """
        super().load_class(klass)
        if isinstance(klass, ComponentBase):
            try:
                key = self.get_instance_registry_key(klass)
                if key:
                    if key not in self.instances.keys():
                        self.instances[key] = klass
                    
            except Exception:
                if self.show_errors:
                    klass_type = self.get_type(klass)
                    message = "Unable to register instance {0} with type {1}".format(klass.__name__, klass_type)
                    logger.error(message)

    def get_instance(self, key):
        self.load_deferred_modules()
        return self.instances.get(key, None)

    def unregister(self, klass):
        """
        Removes a class or instance from the registry
        """
        super().unregister(klass)
        key = self.get_instance_registry_key(klass)
        if key:
            self.instances.pop(key)

    def find_instance(self, string_with_key_or_name) -> typing.List[str]:
        is_key = self.instance_regex.match(string_with_key_or_name)
        if is_key:
            return [self.get_instance(string_with_key_or_name)]
        else:
            self.load_deferred_modules()
            return [x for x in self.instances.keys() if string_with_key_or_name in x]



import os
from environs import Env

env = Env()
env.read_env()

PRESALYTICS = {
    'USERNAME': os.environ['PRESALYTICS_USERNAME'],
    'PASSWORD': os.environ['PRESALYTICS_PASSWORD'],
    'CLIENT_ID': os.environ.get('CLIENT_ID'),
    'CLIENT_SECRET': os.environ.get('CLIENT_SECRET'),
    'HOSTS': {},
    'BROWSER_API_HOST': {}
}

try:
    ooxml_host = {
        'OOXML_AUTOMATION': os.environ['OOXML_AUTOMATION_HOST']
    }
    PRESALYTICS['HOSTS'].update(ooxml_host)
except KeyError:
    pass

try:
    story_host = {
        'STORY': os.environ['STORY_HOST'] + "/story"
    }
    PRESALYTICS['HOSTS'].update(story_host)
except KeyError:
    pass


try:
    site_host = {
        'SITE': os.environ['SITE_HOST']
    }
    PRESALYTICS['HOSTS'].update(site_host)
except KeyError:
    pass

try:
    PRESALYTICS['BROWSER_API_HOST']['STORY'] = os.environ['BROWSER_API_HOST_STORY']
except KeyError:
    pass

try:
    PRESALYTICS['BROWSER_API_HOST']['OOXML_AUTOMATION'] = os.environ['BROWSER_API_HOST_OOXML_AUTOMATION']
except KeyError:
    pass




# coding: utf-8

"""
    OOXML Automation

    This API helps users convert Excel and Powerpoint documents into rich, live dashboards and stories.  # noqa: E501

    The version of the OpenAPI document: 0.1.0-no-tags
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six


class ChildObjects(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'entity_id': 'str',
        'entity_name': 'str',
        'object_type': 'str',
        'parent_entity_id': 'str',
        'parent_object_type': 'str'
    }

    attribute_map = {
        'entity_id': 'entityId',
        'entity_name': 'entityName',
        'object_type': 'objectType',
        'parent_entity_id': 'parentEntityId',
        'parent_object_type': 'parentObjectType'
    }

    def __init__(self, entity_id=None, entity_name=None, object_type=None, parent_entity_id=None, parent_object_type=None):  # noqa: E501
        """ChildObjects - a model defined in OpenAPI"""  # noqa: E501

        self._entity_id = None
        self._entity_name = None
        self._object_type = None
        self._parent_entity_id = None
        self._parent_object_type = None
        self.discriminator = None

        self.entity_id = entity_id
        self.entity_name = entity_name
        self.object_type = object_type
        self.parent_entity_id = parent_entity_id
        self.parent_object_type = parent_object_type

    @property
    def entity_id(self):
        """Gets the entity_id of this ChildObjects.  # noqa: E501


        :return: The entity_id of this ChildObjects.  # noqa: E501
        :rtype: str
        """
        return self._entity_id

    @entity_id.setter
    def entity_id(self, entity_id):
        """Sets the entity_id of this ChildObjects.


        :param entity_id: The entity_id of this ChildObjects.  # noqa: E501
        :type: str
        """

        self._entity_id = entity_id

    @property
    def entity_name(self):
        """Gets the entity_name of this ChildObjects.  # noqa: E501


        :return: The entity_name of this ChildObjects.  # noqa: E501
        :rtype: str
        """
        return self._entity_name

    @entity_name.setter
    def entity_name(self, entity_name):
        """Sets the entity_name of this ChildObjects.


        :param entity_name: The entity_name of this ChildObjects.  # noqa: E501
        :type: str
        """

        self._entity_name = entity_name

    @property
    def object_type(self):
        """Gets the object_type of this ChildObjects.  # noqa: E501


        :return: The object_type of this ChildObjects.  # noqa: E501
        :rtype: str
        """
        return self._object_type

    @object_type.setter
    def object_type(self, object_type):
        """Sets the object_type of this ChildObjects.


        :param object_type: The object_type of this ChildObjects.  # noqa: E501
        :type: str
        """

        self._object_type = object_type

    @property
    def parent_entity_id(self):
        """Gets the parent_entity_id of this ChildObjects.  # noqa: E501


        :return: The parent_entity_id of this ChildObjects.  # noqa: E501
        :rtype: str
        """
        return self._parent_entity_id

    @parent_entity_id.setter
    def parent_entity_id(self, parent_entity_id):
        """Sets the parent_entity_id of this ChildObjects.


        :param parent_entity_id: The parent_entity_id of this ChildObjects.  # noqa: E501
        :type: str
        """

        self._parent_entity_id = parent_entity_id

    @property
    def parent_object_type(self):
        """Gets the parent_object_type of this ChildObjects.  # noqa: E501


        :return: The parent_object_type of this ChildObjects.  # noqa: E501
        :rtype: str
        """
        return self._parent_object_type

    @parent_object_type.setter
    def parent_object_type(self, parent_object_type):
        """Sets the parent_object_type of this ChildObjects.


        :param parent_object_type: The parent_object_type of this ChildObjects.  # noqa: E501
        :type: str
        """

        self._parent_object_type = parent_object_type

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ChildObjects):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



"""
Unit test for Client class
"""
import unittest
import os
import uuid
import presalytics.lib.exceptions
import presalytics.client.api


class TestClient(unittest.TestCase):
    """ Test whether client accurately creates a configuration
        and establishes a connection to the server
    """
    def setUp(self):
        self.config_file = os.path.join(os.path.dirname(__file__), 'files', 'config.py')

    def test_client(self):
        """
        Tests for client configuration into presalytics api via device grant
        """
        if not presalytics.CONFIG.get("PASSWORD", None):
            client = presalytics.client.api.Client(config_file=self.config_file)
            username = os.environ["PRESALYTICS_USERNAME"]
            self.assertEqual(client.username, username)
            self.assertEqual(client.ooxml_automation.api_client.configuration.host, os.environ["OOXML_AUTOMATION_HOST"])
            self.assertNotEqual(client.token_util.token.get("access_token", None), None)
       
    def test_password_grant(self):
        """
        tests if password grant works
        """
        if presalytics.CONFIG.get("PASSWORD", None):
            username = presalytics.CONFIG.get("USERNAME")
            password = presalytics.CONFIG.get("PASSWORD")
            client_id = os.environ.get("CLIENT_ID")
            client_secret = os.environ.get("CLIENT_SECRET")
            client = presalytics.client.api.Client(
                username=username, 
                password=password, 
                client_id=client_id, 
                client_secret=client_secret
            )

    def test_module_config(self):
        from test.files.config import PRESALYTICS
        client = presalytics.client.api.Client(config=PRESALYTICS)
        self.assertIsNotNone(client.client_id)

    def test_api_exception(self):
        exception_client = presalytics.client.api.Client(config_file=self.config_file)
        bad_story_id = str(uuid.uuid4())
        try:
            body, status, headers = exception_client.story.story_id_get(bad_story_id)
        except Exception as e:
            self.assertEqual(presalytics.lib.exceptions.ApiException, e.__class__)
        ignore_client = presalytics.client.api.Client(config_file=self.config_file, ignore_api_exceptions=True)
        body, status, headers = ignore_client.story.story_id_get(bad_story_id)
        self.assertEqual(status, 404)

    def test_file_download(self):
        test_upload_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'files', 'star.pptx')
        client = presalytics.client.api.get_client()
        try:
            test_story = client.upload_file_and_await_outline(test_upload_file)
            download_folder = os.path.dirname(os.path.abspath(__file__))
            test_filename = "test-file.pptx"
            client.download_file(test_story.id, test_story.ooxml_documents[0].ooxml_automation_id, filename=test_filename, download_folder=download_folder)
            target_path = os.path.join(download_folder, test_filename)
            self.assertTrue(os.path.exists(target_path))
        finally:
            try:
                os.remove(target_path)
            except Exception:
                pass

    def test_clone(self):
        test_upload_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'files', 'star.pptx')
        client = presalytics.client.api.get_client()
        try:
            test_story = client.upload_file_and_await_outline(test_upload_file)
            from presalytics.client.presalytics_ooxml_automation.models import DocumentCloneDTO
            ooxml_id = test_story.ooxml_documents[0].ooxml_automation_id
            clone_dto = DocumentCloneDTO(
                id=ooxml_id,
                story_id=str(uuid.uuid4())
            )
            cloned_document = client.ooxml_automation.documents_clone_post_id(ooxml_id, document_clone_dto=clone_dto)

            new_story = client.story.story_post({'outline': test_story.outline})


        finally:
            try:
                client.story.story_id_delete(test_story.id)
            except Exception:
                pass
            try:
                client.ooxml_automation.documents_delete_id(cloned_document.id)
            except Exception:
                pass
            # try:
            #     client.story.story_id_delete(new_story.id)
            # except Exception:
            #     pass




import datetime
import typing
import logging
import presalytics
import presalytics.story.outline
import presalytics.client.api
import presalytics.lib.widgets.ooxml
import presalytics.lib.themes.ooxml

if typing.TYPE_CHECKING:
    from presalytics.client.presalytics_story import Story
    from presalytics.client.presalytics_ooxml_automation import Document
    from presalytics.story.components import ThemeBase
    from presalytics.story.outline import StoryOutline


logger = logging.getLogger(__name__)


def add_ooxml_document_to_story(story_id, new_document_filepath, replace_id=None, username=None, password=None):
    """
    Utility method for asscoiating a new Document in the Presalytics API Ooxml Automation service
    with a Story.  

    Parameters
    ----------
    story_id: str
        The Presalytics API Story service Id for for the story

    new_document_filepath str
        The local filepath to the document tha you want to associate with the story

    replace_id: str, option
        If you want to replace a document, this the documents Ooxml Automation service Id  
        
        Passing this value will also update the references to Ooxml Automation
        service Document object in the Story Outline.  This is a good option if you have
        not made significant changes to the new version of the document.  Widgets may 
        fail to render if more than minor changes are made to the object tree.

    username: str, option
        Presalytics API username.  Defaults to workspace username.

    password: str, option
        Presalytics API password. 
    """

    client = presalytics.client.api.Client(username=username, password=password)
    if story_id == "empty" or not story_id:
        message = "The story_id must not be empty"
        raise presalytics.lib.exceptions.InvalidArgumentException(message)
    replace = True if replace_id else False
    story = client.story.story_id_file_post(story_id, file=new_document_filepath, replace_existing=replace, obsolete_id=replace_id)

def create_story_from_ooxml_file(filename: str, client_info={}) -> 'Story':
    """
    Utility Method for building stories into the Presalytics API directly from a Presentation or Spreadsheet file. 

    Parameters
    ----------
    filename : str
        A string contain the local path to a presenation or spreadsheet object.

    client_info : dict, optional
        A dictionary containing arguments that will be unpacked and passed to a `presalytics.client.api.Client` object on intialization. 
        This dictionary can include the `token`, `cache_tokens`, `delegate_login` values.  See `presalytics.client.api.Client` for more information.

    Returns:
    ----------
    A `presalytics.client.presalytics_story.models.story.Story` containing information about the Story object in the Presalytics API

    """

    story: 'Story'
    logger.info("Starting presalytics tool: create_story_from_ooxml_file")
    logger.info("Intializing presalytics client.")
    
    client = presalytics.Client(**client_info)
    logger.info("Sending file to presalytics server for document processing and base story creation")
    story = client.story.story_post_file(file=filename)
    logger.info("Creating local instances of file widgets")
    outline = presalytics.StoryOutline.load(story.outline)
    for i in range(0, len(outline.pages)):
        page = outline.pages[i]
        for j in range(0, len(page.widgets)):
            widget = page.widgets[j]
            logger.info('Creating OoxmlFileWidget with name "{}"'.format(widget.name))
            inst = presalytics.OoxmlFileWidget.deserialize(widget, **client_info)
            presalytics.COMPONENTS.register(inst)
            logger.info('Rewriting outline with widget: "{}"'.format(widget.name))
            outline.pages[i].widgets[j] = inst.serialize()
    story.outline = outline.dump()
    return story


def create_outline_from_ooxml_document(story_api: 'Story',
                                       ooxml_document: 'Document',
                                       title: str = None,
                                       description: str = None,
                                       themes: typing.Sequence['ThemeBase'] = None,
                                       client_info={}):
    """
    Utility Method for generating a story from a presenation or spreadsheet file

    This method encapsulates a standard workflow for interating the the Presalytics
    API into a single function.  This method takes known `presalytics.client.presalytics_ooxml_automation.models.document.Document`
    with a known `presalytics.client.presalytics_story.models.story.Story`.  This method returns a `presalytics.story.outline.StoryOutline`
    that can be used to replace with existing `presalytics.story.outline.StoryOutline` on the story if the user warrants it.
    

    Parameters
    ----------
    story_api : presalytics.client.presalytics_story.models.story.Story
        The represenation for the Presalytics API instance that will recieve a new associate with the  **ooxml_document**

    ooxml_document : presalytics.client.presalytics_ooxml_automation.models.document.Document 
        the new ooxml_document will be associated with the story.  The outline will be generated from this object

    title : str, optional
        If not provided, the story will take on the title of the `presalytics.client.presalytics_ooxml_automation.models.document.Document` object

    description : str, optional
        If not provided, the story description will initialize as the empty string
    
    themes : list of presalytics.story.components.ThemeBase, optional
        A list of the themes to added to the returned `presalytics.story.outline.StoryOutline`
    
    client_info : dict, optional
        A dictionary containing arguments that will be unpacked and passed to a `presalytics.client.api.Client` object on intialization. 
        This dictionary can include the `token`, `cache_tokens`, `delegate_login` values.  See `presalytics.client.api.Client` for more information.

    Returns
    ----------
    A `presalytics.story.outline.StoryOutline` for the user to optionally serialize and append to the story in downstream operations

    """
    info = presalytics.story.outline.Info(
        revision=0,
        date_created=datetime.datetime.utcnow().isoformat(),
        date_modified=datetime.datetime.utcnow().isoformat(),
        created_by=presalytics.CONFIG.get("USERNAME", ""),
        modified_by=presalytics.CONFIG.get("USERNAME", ""),
        revision_notes='Created by via "create_outline_from_ooxml_file" method'
    )

    if description:
        _description = description
    else:
        _description = ""
    try:
        pages = create_pages_from_ooxml_document(story_api, ooxml_document, client_info=client_info)
    except Exception as ex:
        logger.error("Error adding pages to outline", exc_info=True)
        pages = []
    try:
        if themes:
            _themes = themes
        else:
            ooxml_id = pages[0].widgets[0].data["document_ooxml_id"]
            _themes = [create_theme_from_ooxml_document(ooxml_id, client_info=client_info)]
    except Exception:
        logger.error("Unable to add theme to ooxml_story", exc_info=True)
        _themes = []
    
    if title:
        _title = title
    else:
        _title = pages[0].widgets[0].name

    outline = presalytics.story.outline.StoryOutline(
        info=info,
        pages=pages,
        description=_description,
        title=_title,
        themes=_themes,
        plugins=[]
    )
    return outline


def create_pages_from_ooxml_document(story: 'Story', 
                                     ooxml_document: 'Document',
                                     client_info={}):
    """
    Utility Method for building stories into the Presalytics API directly from a Presenation or Spreadsheet file. 

    Parameters
    ----------
    story : presalytics.client.presalytics_story.models.story.Story
        Representation for Presalytics API Story object

    ooxml_document: presalytics.client.presalytics_ooxml_automation.models.document.Document
        The ooxml_document on the story that you want to create pages from

    client_info : dict, optional
        A dictionary containing arguments that will be unpacked and passed to a `presalytics.client.api.Client` object on intialization. 
        This dictionary can include the `token`, `cache_tokens`, `delegate_login` values.  See `presalytics.client.api.Client` for more information.

    Returns:
    ----------
    A list of `presalytics.story.outline.Page` objects representing the slides, pages and charts in the source document

    """
    pages = []
    order = []
    pages_unordered = []
    client = presalytics.Client(**client_info)
    child_objects = client.ooxml_automation.documents_childobjects_get_id(ooxml_document.id)
    document_type = client.ooxml_automation.documents_documenttype_typeid_get_type_id(ooxml_document.document_type_id)
    if document_type.file_extension == "pptx":
        slides_meta = [x for x in child_objects if x.object_type == "Slide.Slides"]
        ep_map = presalytics.OoxmlEndpointMap.slide()
        for slide in slides_meta:
            try:
                widget = presalytics.OoxmlFileWidget(
                    filename=ooxml_document.filename,
                    name=slide.entity_name,
                    endpoint_map=ep_map,
                    object_name=slide.entity_name,
                    object_ooxml_id=slide.entity_id,
                    document_ooxml_id=ooxml_document.id,
                    story_id=story.id,
                    client_info=client_info
                )
                widget_kind = "widget-page"
                widget_name = slide.entity_name
                widgets = [widget.serialize()]
                page = presalytics.story.outline.Page(
                    kind=widget_kind,
                    name=widget_name,
                    widgets=widgets
                )

                this_slide_meta = client.ooxml_automation.slides_slides_get_id(slide.entity_id)
                order.append(this_slide_meta.number -1)
                pages_unordered.append(page)
            except:
                logger.error("Unable to add widget {0} to outline ooxml document {1}".format(slide.entity_name, ooxml_document.id))
    # TODO: insert excel chart handling here
    for j in range(0, len(order)):
        idx = order.index(j)
        pages.append(pages_unordered[idx])    
    return pages

def create_theme_from_ooxml_document(document_id: str, client_info={}):
    """
    Creates a `presalytics.story.outline.Theme` object from an Presalytics API ooxml_document object

    Parameters:
    ----------
    document_id: str
        A string containing a uuid that corresponds to a document in the Ooxml Automation service of the Presalytics API

    client_info : dict, optional
        A dictionary containing arguments that will be unpacked and passed to a `presalytics.client.api.Client` object on intialization. 
        This dictionary can include the `token`, `cache_tokens`, `delegate_login` values.  See `presalytics.client.api.Client` for more information.

    Returns:
    ----------
    A `presalytics.story.outline.Theme` object with formats extracted from the ooxml_document

    """
    client = presalytics.Client(**client_info)
    child_objects = client.ooxml_automation.documents_childobjects_get_id(document_id)
    themes = [x for x in child_objects if x.object_type == "Theme.Themes"]
    if len(themes) > 1:
        slide_no = None
        for theme_data in themes:
            theme_info = client.ooxml_automation.theme_themes_get_id(theme_data.entity_id)
            parent_slide = client.ooxml_automation.slides_slides_get_id(theme_info.slide_id)
            if slide_no is None:
                slide_no = parent_slide.number
                theme_meta = theme_info
            else:
                if parent_slide.number < slide_no:
                    theme_meta = theme_info
                    if parent_slide.number == 0:
                        break
    else:
        theme_meta = client.ooxml_automation.theme_themes_get_id(themes[0].entity_id)
    theme = presalytics.lib.themes.ooxml.OoxmlTheme(
        theme_meta.name, 
        theme_meta.id, 
        client_info=client_info
    )
    return theme.serialize().to_dict()

def get_mime_type_from_filename(client: presalytics.Client, filename) -> typing.Optional[str]:
    """
    Determines the mimetype from a file's type extension

    Parameters:
    ----------
    client : presalytics.client.api.Client
        A client object for making api calls
    
    filename : str
        A filename against which the method can lookup mimetypes

    Returns:
    ----------
    A string containing a mimetype to attach to file uploads

    """
    doc_types = client.ooxml_automation.documents_documenttype_get()
    file_extension = filename.split(".")[-1]
    try:
        return next(x.mime_type for x in doc_types if x.file_extension == file_extension)
    except StopIteration:
        return None
        